
import warnings
import numpy as np
import pandas as pd

import pymc3 as pm
import theano.tensor as tt

from gumbi.utils.misc import assert_in, assert_is_subset
from gumbi.utils.gp_utils import get_ℓ_prior
from gumbi.aggregation import DataSet
from gumbi.arrays import *
from gumbi.arrays import ParameterArray as parray
from gumbi.arrays import UncertainParameterArray as uparray
from gumbi.arrays import MVUncertainParameterArray as mvuparray

from .base import Regressor

__all__ = ['GP']

class GP(Regressor):
    r"""Gaussian Process surface learning and prediction.

    See Also
    --------
    :class:`Regressor`

    Notes
    -----
    The GP class is built from a dataframe in the form of a :class:`DataSet` object. The output(s) are
    taken from :attr:`DataSet.outputs` and the corresponding column in the tidy data frame taken from
    :attr:`DataSet.names_column`. This column will be generically referred to as the `output_column` in this documentation,
    but can take any value specifed when the :class:`DataSet` is constructed. The model inputs are constructed by
    filtering this dataframe, extracting column values, and converting these to numerical input coordinates. The main
    entry point will be :meth:`fit`, which parses the dimensions of the model with :meth:`specify_model`,
    extracts numerical input coordinates with :meth:`get_structured_data`, compiles the Pymc3 model with
    :meth:`build_model`, and finally learns the hyperparameters with :meth:`find_MAP`.

    Dimensions fall into several categories:

    * Filter dimensions, those with only one level, are used to subset the dataframe but are not included as explicit
      inputs to the model. These are not specified explicitly, but rather any continuous or categorical dimension with only
      one level is treated as a filter dimension.

    * Continuous dimensions are treated as explicit coordinates and given a Radial Basis Function kernel

      * Linear dimensions (which must be a subset of `continuous_dims`) have an additional linear kernel.

    * Coregion dimensions imply a distinct but correlated output for each level

      * If more than one output is specified, the `output_column` is treated as a categorical dim.

    A non-additive model has the form:

    .. math::

        y &\sim \text{Normal} \left( \mu, \sigma \right) \\
        mu &\sim \mathcal{GP} \left( K \right) \\
        K &= \left( K^\text{cont}+K^\text{lin} \right) K^\text{coreg}_\text{outputs} \prod_{n} K^\text{coreg}_{n} \\
        K^\text{cont} &= \text{RBF} \left( \ell_{i}, \eta \right) \\
        K^\text{lin} &= \text{LIN} \left( c_{j}, \tau \right) \\
        K^\text{coreg} &= \text{Coreg} \left( \boldsymbol{W}, \kappa \right) \\
        \sigma &\sim \text{Exponential} \left( 1 \right) \\

    Where :math:`i` denotes a continuous dimension, :math:`j` denotes a linear dimension, and :math:`n` denotes a
    categorical dimension (excluding the `output_column`). :math:`K^\text{cont}` and :math:`K^\text{lin}` each consist of a
    joint kernel encompassing all continuous and linear dimensions, respectively, whereas :math:`K^\text{coreg}_{n}` is
    a distinct kernel for a given categorical dimension.

    The additive model has the form:

    .. math::

        mu &\sim \mathcal{GP}\left( K^\text{global} \right) + \sum_{n} \mathcal{GP}\left( K_{n} \right) \\
        K^\text{global} &= \left( K^\text{cont}+K^\text{lin} \right) K^\text{coreg}_\text{outputs} \\
        K_{n} &= \left( K^\text{cont}_{n}+K^\text{lin}_{n} \right) K^\text{coreg}_\text{outputs} K^\text{coreg}_{n} \\

    Note that, in the additive model, :math:`K^\text{cont}_{n}` and :math:`K^\text{lin}_{n}` still consist of only
    the continuous and linear dimensions, respectively, but have unique priors corresponding to each categorical dimension.
    However, there is only one :math:`K^\text{coreg}_\text{outputs}` kernel.

    Parameters
    ----------
    dataset : DataSet
        Data for fitting.
    outputs : str or list of str, default None
        Name(s) of output(s) to learn. If ``None``, uses all values from ``outputs`` attribute of *dataset*.
    seed : int
        Random seed

    Examples
    --------
    A GP object is created from a :class:`DataSet` and can be fit immediately with the default dimension
    configuration (regressing `r` with RBF + linear kernels for `X` and `Y`):

    >>> import gumbi as gmb
    >>> df = pd.read_pickle(gmb.data.example_dataset)
    >>> outputs=['a', 'b', 'c', 'd', 'e', 'f']
    >>> ds = gmb.DataSet(df, outputs=outputs, log_vars=['Y', 'b', 'c', 'd', 'f'], logit_vars=['X', 'e'])
    >>> gp = gmb.GP(ds, outputs='d').fit()

    Note that last line is equivalent to

    >>> gp = gmb.GP(ds, outputs='d')
    >>> gp.specify_model()
    >>> gp.build_model()
    >>> gp.find_MAP()

    The model can be specified with various continuous, linear, and categorical dimensions.
    `X` and `Y` are always included in both ``continuous_dims`` and ``linear_dims``.

    >>> gp.specify_model(continuous_dims='lg10_Z', linear_dims='lg10_Z', categorical_dims='Pair')
    >>> gmb.GP(ds).fit(continuous_dims='lg10_Z', linear_dims='lg10_Z', categorical_dims='Pair')  # equivalent

    After the model is fit, define a grid of points at which to make predictions. The result is a
    :class:`ParameterArray`:

    >>> gp.prepare_grid()
    >>> gp.grid_points
    ('X', 'Y'): [(0.075     ,  10.) (0.08358586,  10.) (0.09217172,  10.) ...
     (0.90782828, 800.) (0.91641414, 800.) (0.925     , 800.)]

    Make predictions, returning an :class:`UncertainParameterArray`

    >>> gp.predict_grid()
    >>> gp.predictions
    d['μ', 'σ2']: [[(0.70728056, 0.16073197) (0.70728172, 0.16073197)
                    (0.70728502, 0.16073197) ... (0.70727954, 0.16073197)
                    (0.7072809 , 0.16073197) (0.70728058, 0.16073197)]
                   ...
                   [(0.70749247, 0.1607318 ) (0.70773573, 0.16073116)
                    (0.70806603, 0.16072949) ... (0.70728449, 0.16073197)
                    (0.70728194, 0.16073197) (0.7072807 , 0.16073197)]]

    The `uparray` makes it easy to calculate standard statistics in natural or transformed/standardized space while
    maintaining the original shape of the array:

    >>> gp.predictions.z.dist.ppf(0.025)
    array([[-3.1887916 , -3.18878491, -3.18876601, ..., -3.18879744,
            -3.18878966, -3.18879146],
           ...,
           [-3.1875742 , -3.18617286, -3.18426272, ..., -3.18876906,
            -3.18878366, -3.18879081]])

    Finally, plot the results:

    >>> import matplotlib.pyplot as plt
    >>>
    >>> plt.style.use(str(gmb.style.futura))
    >>> x_pa = gp.predictions_X
    >>> y_upa = gp.predictions
    >>> gmb.ParrayPlotter(x_pa, y_upa).plot()

    Plot a slice down the center of the prediction along each axis

    >>> x_pa, y_upa = gp.get_conditional_prediction(Y=88)
    >>>
    >>> ax = gmb.ParrayPlotter(x_pa, y_upa).plot()
    >>> ax.set_xticklabels([int(float(txt.get_text())*100) for txt in ax.get_xticklabels()]);

    Plot a slice down the center of the prediction along each axis

    >>> x_pa, y_upa = gp.get_conditional_prediction(X=0.5)
    >>>
    >>> ax = gmb.ParrayPlotter(x_pa, y_upa).plot()
    >>> ax.set_xticklabels([int(float(txt.get_text())*100) for txt in ax.get_xticklabels()]);


    Attributes
    ----------
    dataset : DataSet
        Data for fitting.
    outputs : list of str
        Name(s) of output(s) to learn.
    seed : int
        Random seed
    continuous_dims : list of str
        Columns of dataframe used as continuous dimensions
    linear_dims : list of str
        Subset of continuous dimensions to apply an additional linear kernel.
    continuous_levels : dict
        Values considered within each continuous column as ``{dim: [level1, level2]}``
    continuous_coords : dict
        Numerical coordinates of each continuous level within each continuous dimension as ``{dim: {level: coord}}``
    categorical_dims : list of str
        Columns of dataframe used as categorical dimensions
    categorical_levels : dict
        Values considered within each categorical column as ``{dim: [level1, level2]}``
    categorical_coords : dict
        Numerical coordinates of each categorical level within each categorical dimension as ``{dim: {level: coord}}``
    additive : bool
        Whether to treat categorical dimensions as additive or joint
    filter_dims : dict
        Dictionary of column-value pairs used to filter dataset before fitting
    X : array
        A 2D tall array of input coordinates.
    y : array
        A 1D vector of observations
    model : pymc3.model.Model
        Compiled pymc3 model
    gp_dict : dict
        Dictionary of model GP objects. Contains at least 'total'.
    """

    def __init__(self, dataset: DataSet, outputs=None, seed=2021):
        super(GP, self).__init__(dataset, outputs, seed)

        self.model = None
        self.gp_dict = None
        self.MAP = None

        self.continuous_kernel = 'ExpQuad'
        self.heteroskedastic_inputs = False
        self.heteroskedastic_outputs = True
        self.sparse = False
        self.n_u = 100

        self.model_specs = {
            'seed': self.seed,
            'continuous_kernel': self.continuous_kernel,
            'heteroskedastic_inputs': self.heteroskedastic_inputs,
            'heteroskedastic_outputs': self.heteroskedastic_outputs,
            'sparse': self.sparse,
            'n_u': self.n_u,
        }

    ################################################################################
    # Model building and fitting
    ################################################################################

    def fit(self, outputs=None, linear_dims=None, continuous_dims=None, continuous_levels=None, continuous_coords=None,
            categorical_dims=None, categorical_levels=None, additive=False, seed=None, heteroskedastic_inputs=False,
            heteroskedastic_outputs=True, sparse=False, n_u=100, **MAP_kwargs):
        """Fits a GP surface

        Parses inputs, compiles a Pymc3 model, then finds the MAP value for the hyperparameters. `{}_dims` arguments
        indicate the columns of the dataframe to be included in the model, with `{}_levels` indicating which values of
        those columns are to be included (``None`` implies all values).

        If ``additive==True``, the model is constructed as the sum of a global GP and a distinct GP for each categorical
        dimension. Each of these GPs, including the global GP, consists of an RBF+linear kernel multiplied by a
        coregion kernel for the `output_column` if necessary. Although the same continuous kernel structure is used for each
        GP in this model, unique priors are assigned to each distinct kernel. However, there is always only one
        coregion kernel for the `output_column`. The kernel for each dimension-specific GP is further multiplied by a
        coregion kernel that provides an output for each level in that dimension.

        See Also
        --------
        :meth:`build_model`

        Parameters
        ----------
        outputs : str or list of str, default None
            Name(s) of output(s) to learn. If ``None``, :attr:`outputs` is used.
        linear_dims : str or list of str, optional
            Subset of continuous dimensions to apply an additional linear kernel. If ``None``, defaults to ``['Y','X']``.
        continuous_dims : str or list of str, optional
            Columns of dataframe used as continuous dimensions.
        continuous_levels : str, list, or dict, optional
            Values considered within each continuous column as ``{dim: [level1, level2]}``.
        continuous_coords : list or dict, optional
            Numerical coordinates of each continuous level within each continuous dimension as ``{dim: {level: coord}}``.
        categorical_dims : str or list of str, optional
            Columns of dataframe used as categorical dimensions.
        categorical_levels : str, list, or dict, optional
            Values considered within each categorical column as ``{dim: [level1, level2]}``.
        additive : bool, default False
            Whether to treat categorical_dims as additive or joint (default).
        seed : int, optional.
            Random seed for model instantiation. If ``None``, :attr:`seed` is used.
        heteroskedastic_inputs: bool, default False
            Whether to allow heteroskedasticity along continuous dimensions (input-dependent noise)
        heteroskedastic_outputs: bool, default True
            Whether to allow heteroskedasticity between multiple outputs (output-dependent noise). `Not yet implemented`
        sparse: bool, default False
            Whether to use a `sparse approximation`_ to the GP.
        n_u: int, default 100
            Number of inducing points to use for the sparse approximation, if required.
        **MAP_kwargs
            Additional keyword arguments passed to :func:`pm.find_MAP`.

        Returns
        -------
        self : :class:`GP`
        """

        self.specify_model(outputs=outputs, linear_dims=linear_dims, continuous_dims=continuous_dims,
                           continuous_levels=continuous_levels, continuous_coords=continuous_coords,
                           categorical_dims=categorical_dims, categorical_levels=categorical_levels,
                           additive=additive)

        self.build_model(seed=seed,
                         heteroskedastic_inputs=heteroskedastic_inputs,
                         heteroskedastic_outputs=heteroskedastic_outputs,
                         sparse=sparse, n_u=n_u)

        self.find_MAP(**MAP_kwargs)

        return self

    # TODO: add full probabilistic model description to docstring
    # TODO: allow dimension-specific continuous kernel specification
    # TODO: allow single multi-dimensional continuous kernel rather than independent kernels per dimension
    def build_model(self, seed=None, continuous_kernel='ExpQuad', heteroskedastic_inputs=False, heteroskedastic_outputs=True, sparse=False, n_u=100):
        r"""Compile a pymc3 model for the GP.

        Each dimension in :attr:`continuous_dims` is combined in an ExpQuad kernel with a principled
        :math:`\text{InverseGamma}` prior for each lengthscale (as `suggested by Michael Betancourt`_) and a
        :math:`\text{Gamma}\left(2, 1\right)` prior for variance.

        .. _suggested by Michael Betancourt: https://betanalpha.github.io/assets/case_studies/gp_part3/part3.html#4_adding_an_informative_prior_for_the_length_scale
        .. _pymc3 docs: https://docs.pymc.io/en/v3/api/gp/cov.html
        .. _sparse approximation: https://docs.pymc.io/en/v3/pymc-examples/examples/gaussian_processes/GP-SparseApprox.html

        Parameters
        ----------
        seed : int, optional.
            Random seed. If ``None``, :attr:`seed` is used.
        continuous_kernel : {'ExpQuad', 'RatQuad', 'Matern32', 'Matern52', 'Exponential', or 'Cosine'}
            Covariance function to use for continuous dimensions. See `pymc3 docs`_ for more details.
        heteroskedastic_inputs: bool, default False
            Whether to allow heteroskedasticity along continuous dimensions (input-dependent noise).
        heteroskedastic_outputs: bool, default True
            Whether to allow heteroskedasticity between multiple outputs (output-dependent noise). `Not yet implemented`.
        sparse: bool, default False
            Whether to use a `sparse approximation`_ to the GP.
        n_u: int, default 100
            Number of inducing points to use for the sparse approximation, if required.

        Returns
        -------
        self : :class:`GP`
        """

        if heteroskedastic_inputs:
            raise NotImplementedError('Heteroskedasticity over inputs is not yet implemented.')

        X, y = self.get_shaped_data('mean')

        seed = self.seed if seed is None else seed
        self.seed = seed
        self.continuous_kernel = continuous_kernel
        self.heteroskedastic_inputs = heteroskedastic_inputs
        self.heteroskedastic_outputs = heteroskedastic_outputs
        self.sparse = sparse
        self.n_u = n_u

        self.model_specs = {
            'seed': seed,
            'continuous_kernel': continuous_kernel,
            'heteroskedastic_inputs': heteroskedastic_inputs,
            'heteroskedastic_outputs': heteroskedastic_outputs,
            'sparse': sparse,
            'n_u': n_u,
        }

        n_l = len(self.linear_dims)
        n_s = len(self.continuous_dims)
        n_c = len(self.categorical_dims)
        n_p = len(self.outputs)

        D_in = len(self.dims)
        assert X.shape[1] == D_in

        idx_l = [self.dims.index(dim) for dim in self.linear_dims]
        idx_s = [self.dims.index(dim) for dim in self.continuous_dims]
        idx_c = [self.dims.index(dim) for dim in self.categorical_dims]
        idx_p = self.dims.index(self.out_col) if self.out_col in self.dims else None

        X_s = X[:, idx_s]

        ℓ_μ, ℓ_σ = [stat for stat in np.array([get_ℓ_prior(dim) for dim in X_s.T]).T]

        continuous_kernels = ['ExpQuad', 'RatQuad', 'Matern32', 'Matern52', 'Exponential', 'Cosine']
        assert_in('Continuous kernel', continuous_kernel, continuous_kernels)
        continuous_cov_func = getattr(pm.gp.cov, continuous_kernel)

        def continuous_cov(suffix):
            ℓ = pm.InverseGamma(f'ℓ_{suffix}', mu=ℓ_μ, sigma=ℓ_σ, shape=n_s)
            η = pm.Gamma(f'η_{suffix}', alpha=2, beta=1)
            return η ** 2 * continuous_cov_func(input_dim=D_in, active_dims=idx_s, ls=ℓ)

        def linear_cov(suffix):
            c = pm.Normal(f'c_{suffix}', mu=0, sigma=10, shape=n_l)
            τ = pm.HalfNormal(f'τ_{suffix}', sigma=10)
            return τ * pm.gp.cov.Linear(input_dim=D_in, c=c, active_dims=idx_l)

        def coreg_cov(suffix, D_out, idx):
            testval = np.random.default_rng(seed).standard_normal(size=(D_out, 2))
            W = pm.Normal(f"W_{suffix}", mu=0, sd=3, shape=(D_out, 2), testval=testval)
            κ = pm.Gamma(f"κ_{suffix}", alpha=1.5, beta=1, shape=(D_out,))
            return pm.gp.cov.Coregion(input_dim=D_in, active_dims=[idx], kappa=κ, W=W)

        if sparse:
            pm_gp = pm.gp.MarginalSparse
            gp_kws = {'approx': "FITC"}
        else:
            pm_gp = pm.gp.Marginal
            gp_kws = {}


        with pm.Model() as self.model:
            # μ = pm.Normal('μ', mu=0, sigma=10)
            # β = pm.Normal('β', mu=0, sigma=10, shape=n_l)
            # lin_mean = pm.gp.mean.Linear(coeffs=[β[i] if i in idx_l else 0 for i in range(D_in), intercept=μ)

            # Define a "global" continuous kernel regardless of additive structure
            cov = continuous_cov('total')
            if n_l > 0:
                cov += linear_cov('total')

            # Construct a coregion kernel for each categorical_dims
            if n_c > 0 and not self.additive:
                for dim, idx in zip(self.categorical_dims, idx_c):
                    if dim == self.out_col:
                        continue
                    D_out = len(self.categorical_levels[dim])
                    cov *= coreg_cov(dim, D_out, idx)

            # Coregion kernel for parameters, if necessary
            if self.out_col in self.categorical_dims:
                D_out = len(self.categorical_levels[self.out_col])
                cov_param = coreg_cov(self.out_col, D_out, idx_p)
                cov *= cov_param

            gp_dict = {'total': pm_gp(cov_func=cov, **gp_kws)}

            # Construct a continuous+coregion kernel for each categorical_dim, then combine them additively
            if self.additive:
                gp_dict['global'] = gp_dict['total']
                for dim, idx in zip(self.categorical_dims, idx_c):
                    if dim == self.out_col:
                        continue

                    # Continuous kernel specific to this dimension
                    cov = continuous_cov(dim)
                    # TODO: Decide if each additive dimension needs its own linear kernel
                    if n_l > 0:
                        cov += linear_cov(dim)

                    # Coregion kernel specific to this dimension
                    D_out = len(self.categorical_levels[dim])
                    cov *= coreg_cov(dim, D_out, idx)

                    # Coregion kernel for parameters, if necessary
                    if self.out_col in self.categorical_dims:
                        cov *= cov_param

                    # Combine GPs
                    gp_dict[dim] = pm_gp(cov_func=cov, **gp_kws)
                    gp_dict['total'] += gp_dict[dim]

            # From https://docs.pymc.io/notebooks/GP-Marginal.html
            # OR a covariance function for the noise can be given
            # noise_l = pm.Gamma("noise_l", alpha=2, beta=2)
            # cov_func_noise = pm.gp.cov.Exponential(1, noise_l) + pm.gp.cov.WhiteNoise(sigma=0.1)
            # y_ = gp.marginal_likelihood("y", X=X, y=y, noise=cov_func_noise)


            # GP is heteroskedastic across outputs by default,
            # but homoskedastic across continuous dimensions
            σ = pm.Exponential('σ', lam=1)
            noise = pm.gp.cov.WhiteNoise(sigma=σ)
            if heteroskedastic_inputs:
                raise NotImplementedError('Heteroskedasticity over inputs is not yet implemented')
                # noise += continuous_cov('noise')
            if heteroskedastic_outputs and self.out_col in self.categorical_dims:
                D_out = len(self.categorical_levels[self.out_col])
                noise *= coreg_cov('Output_noise', D_out, idx_p)

            if sparse:
                Xu = pm.gp.util.kmeans_inducing_points(n_u, X)
                if heteroskedastic_outputs:
                    warnings.warn('Heteroskedasticity over outputs is not yet implemented for sparse GP. Reverting to scalar-valued noise.')
                _ = gp_dict['total'].marginal_likelihood('ml', X=X, Xu=Xu, y=y, noise=σ)
            else:
                _ = gp_dict['total'].marginal_likelihood('ml', X=X, y=y, noise=noise)

        self.gp_dict = gp_dict
        return self

    def find_MAP(self, *args, **kwargs):
        """Finds maximum a posteriori value for hyperparameters in model

        Parameters
        ----------
        *args
            Positional arguments passed to :func:`pm.find_MAP`
        **kwargs
            Keyword arguments passed to :func:`pm.find_MAP`
        """
        assert self.model is not None
        with self.model:
            self.MAP = pm.find_MAP(*args, **kwargs)

    def predict(self, points_array, with_noise=True, additive_level='total', **kwargs):
        """Make predictions at supplied points using specified gp

        Parameters
        ----------
        output : str
        points : ParameterArray
            Tall ParameterArray vector of coordinates for prediction, must have one layer per ``self.dims``
        with_noise : bool, default True
            Whether to incorporate aleatoric uncertainty into prediction error

        Returns
        -------
        prediction : UncertainParameterArray
            Predictions as a `uparray`
        """

        # TODO: need to supply "given" dict for additive GP sublevel predictions
        if additive_level != 'total':
            raise NotImplementedError('Prediction for additive sublevels is not yet supported.')

        # Prediction means and variance as a numpy vector
        predictions = self.gp_dict[additive_level].predict(points_array, point=self.MAP, diag=True,
                                                           pred_noise=with_noise, **kwargs)

        return predictions


"""Tests behavior of structured arrays"""

import pytest
import numpy as np
import pathlib as pl

from gumbi import Standardizer
from gumbi import parray, uarray, uparray, mvuparray

test_dir = pl.Path(__file__).resolve().parent
test_data = test_dir / 'test_data'

example_stdzr = {
    'a': {'μ': -0.762, 'σ2': 1.258**2},
    'b': {'μ': -0.0368, 'σ2': 0.351**2},
    'c': {'μ': -5.30, 'σ2': 0.582**2},
    'd': {'μ': -0.307, 'σ2': 0.158**2},
    'e': {'μ': -1.056, 'σ2': 0.398**2},
    'f': {'μ': 3.34, 'σ2': 0.1501**2},
    'X': {'μ': -0.282, 'σ2': 1**2},
    'Y': {'μ': 4.48, 'σ2': 0.75**2},
    'lg10_Z': {'μ': 5, 'σ2': 2**2},
}

log_vars = ['d', 'f', 'b', 'c', 'Y']
logit_vars = ['e', 'X']

def test_parray():
    stdzr = Standardizer(**example_stdzr, log_vars=log_vars, logit_vars=logit_vars)

    # Parameter found in stdzr
    # TODO: Update parray test when change stdzr defaults
    rpa = parray(d=np.arange(5, 10) / 10, stdzr=stdzr)
    assert np.allclose(rpa, np.arange(5, 10) / 10)
    assert np.allclose(rpa.values(), np.arange(5, 10) / 10)
    assert np.allclose(rpa.t, np.array([-0.69314718, -0.51082562, -0.35667494, -0.22314355, -0.10536052]))
    assert np.allclose(rpa.z, np.array([-2.4439695 , -1.29003559, -0.31439838,  0.53073702,  1.27619927]))

    # Composition with numpy functions
    assert np.allclose(np.min(np.sqrt(np.mean(np.square(rpa-rpa[0]-0.05)))).t, -1.5791256)
    assert np.argmax(rpa.values()) == 4

    # Parameter not found in stdzr
    pa1 = parray(param=np.arange(5), stdzr=stdzr)
    assert np.allclose(pa1, np.arange(5))
    assert np.allclose(pa1.t, np.arange(5))
    assert np.allclose(pa1.z, np.arange(5))

    # Multiple parameters
    pa2 = parray(param=np.arange(5), other=np.arange(5)*10, stdzr=stdzr)
    assert np.allclose(pa2.get('param').values(), np.array([0., 1., 2., 3., 4.]))
    assert np.allclose(pa2.get('other').values(), np.array([0., 10., 20., 30., 40.]))
    assert pa2.values().shape == (2, 5)

    # Indexing and slicing
    assert pa1[0].values() == 0
    assert np.allclose(pa1[::2].values(), np.array([0, 2, 4]))
    assert np.allclose(pa2[::2].get('param').values(), np.array([0, 2, 4]))


def test_uarray():
    ua1 = uarray('A', μ=1, σ2=0.1)
    ua2 = uarray('A', μ=2, σ2=0.2)

    # Arithmetic operations
    ua3 = ua1+1
    assert np.isclose(ua3.μ, 2.)
    assert np.isclose(ua3.σ2, 0.1)
    assert np.isclose(ua3.σ, 0.3162277660)

    ua4 = ua2+ua1
    assert np.isclose(ua4.μ, 3.)
    assert np.isclose(ua4.σ2, 0.3)

    ua5 = ua2-ua1
    assert np.isclose(ua5.μ, 1.)
    assert np.isclose(ua5.σ2, 0.3)

    ua6 = uarray.stack([ua1, ua2]).mean(axis=0)
    assert np.isclose(ua6.μ, 1.5)
    assert np.isclose(ua6.σ2, 0.075)

    ua7 = uarray('B', np.arange(1,5)/10, np.arange(1,5)/100)
    assert np.isclose(ua7.mean().μ, 0.25)
    assert np.isclose(ua7.mean().σ2, 0.00625)

    ua8 = ua1+ua7.mean()
    assert ua8.name == '(A+B)'
    assert np.isclose(ua8.μ, 1.25)
    assert np.isclose(ua8.σ2, 0.10625)

    # Conversion to scipy distribution
    assert np.allclose(ua7.dist.ppf(0.95), np.array([0.26448536, 0.43261743, 0.58489701, 0.72897073]))
    rvs = ua7.dist.rvs([3, *ua7.shape], random_state=2021)
    targets = np.array([[0.24886091, 0.29560237, 0.2275221, 0.23869584],
                        [0.15558758, 0.10022663, 0.4958704, 0.52900037],
                        [0.11064137, 0.25970171, 0.32151326, 0.23240931]])
    assert np.allclose(rvs, targets)


def test_uparray():
    stdzr = Standardizer(**example_stdzr, log_vars=log_vars, logit_vars=logit_vars)

    # TODO: Update uparray test when change stdzr defaults
    upa = uparray('c', np.arange(1, 5) / 10, np.arange(1, 5) / 100, stdzr)
    assert np.allclose(upa.μ, np.arange(1, 5) / 10)
    assert np.allclose(upa.σ2, np.arange(1, 5) / 100)

    # Round-trip transformation
    roundtrip_μ, roundtrip_σ2 = upa.stdzr.unstdz(upa.name, upa.z.μ, upa.z.σ2)
    assert np.allclose(upa.μ, roundtrip_μ)
    assert np.allclose(upa.σ2, roundtrip_σ2)

    upa2 = uparray(upa.name, upa.z.μ, upa.z.σ2, stdzr, stdzd=True)
    assert np.allclose(upa.μ, upa2.μ)
    assert np.allclose(upa.σ2, upa2.σ2)
    assert np.allclose(upa.z.μ, upa2.z.μ)
    assert np.allclose(upa.z.σ2, upa2.z.σ2)

    # Correct error propagation
    assert np.isclose(upa.mean().μ, 0.22133638)
    assert np.isclose(upa.mean().σ2, 0.00625)
    analytical_mean = upa.mean().dist.mean()
    mc_mean = np.exp(upa.t.mean().dist.rvs(10000, random_state=2021).mean())
    assert np.isclose(analytical_mean, mc_mean, atol=0.01)
    mc_var = np.log(upa.mean().dist.rvs(10000, random_state=2021)).var()
    assert np.isclose(upa.mean().σ2, mc_var, atol=1e-4)

    # Distribution behavior
    assert np.allclose(upa.dist.ppf(0.025), np.array([0.08220152, 0.1515835 , 0.21364308, 0.27028359]))
    assert np.allclose(upa.dist.ppf(0.975), np.array([0.12165225, 0.26388097, 0.42126336, 0.59197082]))
    rvs = upa.dist.rvs([3, *upa.shape], random_state=2021)
    target = np.array([[0.11605116, 0.22006429, 0.27902589, 0.34041327],
                       [0.10571616, 0.1810085 , 0.36491077, 0.45507622],
                       [0.10106982, 0.21230397, 0.3065239 , 0.33827997]])
    assert np.allclose(rvs, target)


def test_mvuparray():
    stdzr = Standardizer(**example_stdzr, log_vars=log_vars, logit_vars=logit_vars)

    # TODO: Update uparray test when change stdzr defaults
    stdzr = Standardizer(**example_stdzr, log_vars=log_vars, logit_vars=logit_vars)
    c_μ = np.arange(1, 5)/10
    c_σ2 = np.arange(1, 5)/100
    r_μ = c_μ+0.5
    r_σ2 = c_σ2/100*2
    c_upa = uparray('c', c_μ, c_σ2, stdzr)
    r_upa = uparray('d', r_μ, r_σ2, stdzr)
    cor = np.array([[1, -0.6], [-0.6, 1]])
    mvup = mvuparray(c_upa, r_upa, cor=cor)

    # Proper construction
    assert np.allclose(mvup.μ.values(), np.stack([c_μ, r_μ]))
    assert np.allclose(mvup.get('d').μ, r_μ)
    assert np.allclose(mvup.t.get('d_t').μ, r_upa.t.μ)
    assert np.allclose(mvup.t.μ['d_t'], r_upa.t.μ)
    assert np.allclose(mvup.z.μ['d_z'], r_upa.z.μ)

    # Distribution behavior
    pa = mvup.parray(c=0.09, d=0.61)
    assert np.isclose(mvup[0].dist.cdf(pa), 0.0889634198684274)
    assert np.isclose(mvup[0].t.dist.cdf(pa), 0.0889634198684274)
    assert np.isclose(mvup[0].z.dist.cdf(pa), 0.0889634198684274)
    mvup[0].dist.cdf(mvup.μ)  # Should run without errors

    rvs = mvup[0].dist.rvs(2, random_state=2021)
    assert np.allclose(rvs['d'].values(), np.array([0.61310678, 0.59268474]))
    assert np.allclose(rvs['c'].values(), np.array([0.08709257, 0.10308707]))



"""Tests behavior of structured arrays"""

import pytest
import numpy as np
import matplotlib.pyplot as plt
from gumbi import *

def test_1d_array_plotting():
    stdzr = Standardizer(c={'μ': -5.30, 'σ': 0.582}, d={'μ': -0.307, 'σ': 0.158}, z={'μ': 2, 'σ': 2},
                         log_vars=['d', 'c'], logit_vars=['z'])
    m_pa = parray(m=np.arange(1, 5) / 10, stdzr=stdzr)
    r_upa = uparray('d', np.arange(1, 5) / 10 + 0.5, np.arange(1, 5) / 100 * 2, stdzr=stdzr)
    ParrayPlotter(m_pa, r_upa.t).plot()
    ParrayPlotter(m_pa, r_upa.z).plot()
    ParrayPlotter(m_pa.t, r_upa.t).plot()
    ParrayPlotter(m_pa.z, r_upa.z).plot()
    ParrayPlotter(m_pa, r_upa, y_scale='standardized').plot()
    ParrayPlotter(m_pa, r_upa, y_scale='transformed').plot()
    ParrayPlotter(m_pa, r_upa, y_scale='transformed', x_scale='standardized').plot()
    ParrayPlotter(m_pa, r_upa, y_scale='standardized', x_scale='transformed').plot()
    ParrayPlotter(m_pa, r_upa, y_scale='transformed', x_scale='standardized').plot()
    ParrayPlotter(m_pa, r_upa.z, y_tick_scale='standardized').plot()
    ParrayPlotter(m_pa, r_upa.t, y_tick_scale='transformed').plot()
    ParrayPlotter(m_pa.z, r_upa, x_tick_scale='standardized').plot()
    ParrayPlotter(m_pa.t, r_upa, x_tick_scale='transformed').plot()


def test_2d_array_plotting():
    stdzr = Standardizer(c={'μ': -5.30, 'σ': 0.582}, d={'μ': -0.307, 'σ': 0.158}, log_vars=['d', 'c'])
    c = np.arange(1, 10, 0.25)
    d = np.arange(1, 10, 0.25)
    c, d = np.meshgrid(c, d)
    Z = np.sin(np.sqrt((c - 5) ** 2 + (d - 5) ** 2))
    mrz = ParameterArray(c=c, d=d, z=Z, stdzr=stdzr)
    pp = ParrayPlotter(mrz['c'], mrz['d'], mrz['z'])
    cs = pp(plt.contourf)
    pp = ParrayPlotter(mrz['c'].t, mrz['d'].t, mrz['z'].t)
    cs = pp(plt.contourf)
    pp = ParrayPlotter(mrz['c'].z, mrz['d'].z, mrz['z'].z)
    cs = pp(plt.contourf)
    pp = ParrayPlotter(mrz['c'], mrz['d'], mrz['z'],
                       x_scale='standardized', y_scale='standardized', z_scale='standardized')
    cs = pp(plt.contourf)
    pp = ParrayPlotter(mrz['c'], mrz['d'], mrz['z'],
                       x_scale='transformed', y_scale='transformed', z_scale='transformed')
    cs = pp(plt.contourf)
    pp = ParrayPlotter(mrz['c'].z, mrz['d'].t, mrz['z'], x_tick_scale='standardized', y_tick_scale='transformed')
    cs = pp(plt.contourf)
    pp = ParrayPlotter(mrz['c'].t, mrz['d'].z, mrz['z'], x_tick_scale='transformed', y_tick_scale='standardized')
    cs = pp(plt.contourf)

def test_colorbar():
    c = np.arange(1, 10, 0.25)
    d = np.arange(1, 10, 0.25)
    c, d = np.meshgrid(c, d)
    Z = np.sin(np.sqrt((c - 5) ** 2 + (d - 5) ** 2))
    stdzr = Standardizer(c={'μ': -5, 'σ': 0.5}, d={'μ': -0.3, 'σ': 0.15}, z={'μ': 2, 'σ': 2}, log_vars=['d', 'c'],
                         logit_vars=['z'])
    mrz = ParameterArray(c=c, d=d, z=Z ** 2, stdzr=stdzr)

    pp = ParrayPlotter(mrz['c'], mrz['d'], mrz['z'])
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz['c'], mrz['d'], mrz['z'].t)
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz['c'], mrz['d'], mrz['z'].z)
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz['c'], mrz['d'], mrz['z'].t)
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz['c'], mrz['d'], mrz['z'], z_scale='transformed')
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz['c'], mrz['d'], mrz['z'].z)
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz['c'], mrz['d'], mrz['z'], z_scale='standardized')
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)



import pytest
import pickle
import pandas as pd
import numpy as np
import pathlib as pl
import pymc3 as pm

from gumbi import Standardizer, WideData, DataSet

test_dir = pl.Path(__file__).resolve().parent
test_data = test_dir / 'test_data'

example_stdzr = {
    'a': {'μ': -0.762, 'σ2': 1.258**2},
    'b': {'μ': -0.0368, 'σ2': 0.351**2},
    'c': {'μ': -5.30, 'σ2': 0.582**2},
    'd': {'μ': -0.307, 'σ2': 0.158**2},
    'e': {'μ': -1.056, 'σ2': 0.398**2},
    'f': {'μ': 3.34, 'σ2': 0.1501**2},
    'X': {'μ': -0.282, 'σ2': 1**2},
    'Y': {'μ': 4.48, 'σ2': 0.75**2},
    'lg10_Z': {'μ': 5, 'σ2': 2**2},
}

log_vars = ['d', 'f', 'b', 'c', 'Y']
logit_vars = ['e', 'X']

################################################################################
# Standardizer
################################################################################

def test_stdz():
    s = Standardizer(**example_stdzr, log_vars=log_vars, logit_vars=logit_vars)
    nat_defaults = {p: s.untransform(p, v['μ']) for p, v in example_stdzr.items()}
    assert np.allclose([s.stdz(p, nat_defaults[p]) for p, v in nat_defaults.items()], 0), 'Standardization failed'
    assert np.allclose([s.unstdz(p, s.stdz(p, nat_defaults[p])) for p, v in example_stdzr.items()],
                       [v for v in nat_defaults.values()]), 'Round-trip standardization failed'

    stdzr = Standardizer(x={'μ': 1, 'σ2': 0.1}, d={'μ': 0, 'σ2': 0.1}, log_vars=['d'])
    assert stdzr.transform('x', μ=1) == 1
    assert stdzr.stdz('x', 1) == 0.0
    assert stdzr.unstdz('x', 0) == 1.0
    assert np.isclose(stdzr.stdz('x', 1+0.1**0.5), 1.)
    assert np.isclose(stdzr.unstdz('x', 1), 1+0.1**0.5)
    assert stdzr.stdz('d', 1) == 0.0
    assert np.isclose(stdzr.stdz('d', np.exp(0.1**0.5)), 1.0)

    assert stdzr.transform('x', μ=1, σ2=0.1) == (1, 0.1)
    assert stdzr.stdz('x', 1, 0.1) == (0.0, 1.0)
    assert stdzr.stdz('d', 1, 0.1) == (0.0, 1.0)
    assert stdzr.transform('d', 1, 0.1) == (0.0, 0.1)

    x_series = pd.Series(np.arange(1, 5), name='x')
    x_stdzd = np.array([0.0, 3.162278, 6.324555, 9.486833])
    assert np.allclose(stdzr.stdz(x_series).values, x_stdzd)

    r_series = pd.Series(np.arange(1, 5), name='d')
    r_stdzd = np.array([0.0, 2.19192384, 3.4741171 , 4.38384769])
    assert np.allclose(stdzr.stdz(r_series).values, r_stdzd)

################################################################################
# DataSet
################################################################################

@pytest.fixture
def example_dataset():
    df = pd.read_pickle(test_data / 'estimates_test_data.pkl')
    ds = DataSet.from_tidy(df, names_column='Parameter', log_vars=['Y', 'c', 'b'], logit_vars=['X', 'e'])
    assert ds.wide is not None
    assert ds.wide.z is not None
    assert ds.tidy is not None
    assert ds.tidy.z is not None
    return ds


def test_tidy_z(example_dataset):
    ds = example_dataset
    assert ds.tidy.z.shape == ds.tidy.shape
    assert np.allclose([ds.tidy.z[ds.tidy.z.Parameter == p]['Value'].mean() for p in ds.tidy.z.Parameter.unique()], 0)


def test_wide_io(example_dataset):
    ds = example_dataset

    wide_out = ds.wide
    wide_in_wd = WideData(wide_out, outputs=ds.outputs, log_vars=['Y', 'c', 'b'], logit_vars=['X', 'e'])
    wide_in_ds = DataSet(wide_out, outputs=ds.outputs, log_vars=['Y', 'c', 'b'], logit_vars=['X', 'e'])

    pd.testing.assert_frame_equal(wide_in_wd, wide_out)
    pd.testing.assert_frame_equal(wide_in_wd, wide_in_ds.wide)
    pd.testing.assert_frame_equal(ds.wide, wide_in_ds.wide)
    pd.testing.assert_frame_equal(ds.wide, wide_in_wd)

    ds.wide = wide_out.drop(0)
    pd.testing.assert_frame_equal(ds.wide, wide_out.drop(0))
    pd.testing.assert_frame_equal(ds.wide, wide_in_wd.drop(0))




import pathlib as pl
import os

style_dir = pl.Path(__file__).resolve().parent

futura = style_dir / 'futura_presentation.mplstyle'
breve = style_dir / 'breve_presentation.mplstyle'
default = style_dir / 'presentation.mplstyle'


import pytest
import pickle
import pandas as pd
import numpy as np
import pathlib as pl
import pymc3 as pm

from gumbi import GP, DataSet, Standardizer

test_dir = pl.Path(__file__).resolve().parent
test_data = test_dir / 'test_data'

example_stdzr = {
    'a': {'μ': -0.762, 'σ2': 1.258**2},
    'b': {'μ': -0.0368, 'σ2': 0.351**2},
    'c': {'μ': -5.30, 'σ2': 0.582**2},
    'd': {'μ': -0.307, 'σ2': 0.158**2},
    'e': {'μ': -1.056, 'σ2': 0.398**2},
    'f': {'μ': 3.34, 'σ2': 0.1501**2},
    'X': {'μ': -0.282, 'σ2': 1**2},
    'Y': {'μ': 4.48, 'σ2': 0.75**2},
    'lg10_Z': {'μ': 5, 'σ2': 2**2},
}

log_vars = ['d', 'f', 'b', 'c', 'Y']
logit_vars = ['e', 'X']

################################################################################
# GP
################################################################################

@pytest.fixture
def example_estimates():
    es = pd.read_pickle(test_data / 'test_dataset.pkl')
    stdzr = Standardizer(**example_stdzr, log_vars=log_vars, logit_vars=logit_vars)
    ds = DataSet.from_tidy(es, names_column='Parameter', stdzr=stdzr)
    return ds

@pytest.fixture
def example_gp(example_estimates):
    return GP(example_estimates, outputs='d')

# Input argument parsing

def test_gp_default_fit_parsing(example_gp):
    gp = example_gp.specify_model(continuous_dims=['X', 'Y'])
    assert gp.continuous_dims == ['X', 'Y']
    assert gp.categorical_dims == []


def test_gp_data_parsing(example_gp):
    gp = example_gp.specify_model(continuous_dims=['X', 'Y'])
    X, y = gp.get_structured_data()
    assert X.shape == (66,)
    assert len(X.names) == 2
    assert y.shape == (66,)


def test_gp_numerical_continuous_fit_parsing(example_gp):
    gp = example_gp.specify_model(continuous_dims=['X', 'Y', 'lg10_Z'])
    assert gp.continuous_dims == ['X', 'Y', 'lg10_Z']
    for dim in gp.continuous_dims:
        assert len(gp.continuous_levels[dim]) == len(gp.data.tidy[dim].unique())
        assert len(gp.continuous_coords[dim].values()) == len(gp.continuous_levels[dim])
    assert gp.categorical_dims == []
    X, y = gp.get_structured_data()
    assert X.shape == (66,)
    assert len(X.names) == 3
    assert y.shape == (66,)


def test_gp_categorical_continuous_fit_parsing(example_gp):
    gp = example_gp.specify_model(continuous_dims=['X', 'Y', 'Name'])
    assert gp.continuous_dims == ['X', 'Y', 'Name']
    for dim in gp.continuous_dims:
        assert len(gp.continuous_levels[dim]) == len(gp.data.tidy[dim].unique())
        assert len(gp.continuous_coords[dim].values()) == len(gp.continuous_levels[dim])
    assert gp.categorical_dims == []
    X, y = gp.get_structured_data()
    assert X.shape == (66,)
    assert len(X.names) == 3
    assert y.shape == (66,)


def test_gp_params_fit_parsing(example_gp):
    gp = example_gp.specify_model(outputs=['d', 'c'], continuous_dims=['X', 'Y'])
    assert gp.continuous_dims == ['X', 'Y']
    assert gp.categorical_dims == ['Parameter']
    assert gp.categorical_levels == {'Parameter': ['d', 'c']}
    assert gp.categorical_coords == {'Parameter': {'d': 1, 'c': 0}}
    X, y = gp.get_structured_data()
    assert X.shape == (66,)
    assert len(X.names) == 2
    assert y.shape == (66,)
    assert len(y.names) == 2


def test_gp_single_input_fit_parsing(example_gp):
    gp = example_gp.specify_model(continuous_dims=['X', 'Y', 'Name'], continuous_levels={'Name': ['intense-opportunity']})
    assert gp.continuous_dims == ['X', 'Y']
    assert gp.filter_dims == {'Name': ['intense-opportunity'], 'Parameter': ['d']}
    X, y = gp.get_structured_data()
    assert X.shape == (7,)
    assert len(X.names) == 2
    assert y.shape == (7,)

# Model building

def test_gp_build_model_simple(example_gp):
    gp = example_gp.specify_model(continuous_dims=['X', 'Y'])
    gp.build_model()
    assert isinstance(gp.model, pm.model.Model)
    assert isinstance(gp.gp_dict['total'], pm.gp.gp.Marginal)


# Combinatorial gp objects with various parameterizations
@pytest.fixture(params=[False, True])
def additive(request):
    return request.param


@pytest.fixture(params=[{'outputs': ['d', 'c'], 'continuous_dims': ['X', 'Y']},
                        {'continuous_dims': ['X', 'Y'], 'categorical_dims': 'Code'},
                        {'continuous_dims': ['X', 'Y', 'Name']},
                        {'continuous_dims': ['X', 'Y', 'lg10_Z']},
                        {'continuous_dims': ['X', 'Y', 'Name'], 'continuous_levels': {'Name': ['intense-opportunity']}}])
def fit_inputs(request):
    return request.param


@pytest.mark.slow
def test_gp_build_model(example_gp, fit_inputs, additive):
    # Basically just makes sure that `build_model` runs without errors
    gp = example_gp.specify_model(**fit_inputs, additive=additive)
    gp.build_model()
    assert isinstance(gp.model, pm.model.Model)
    assert isinstance(gp.gp_dict['total'], pm.gp.gp.Marginal)


@pytest.mark.slow
def test_gp_build_model_additive(example_gp):
    # Basically just makes sure that `build_model` runs without errors
    gp = example_gp.specify_model(outputs=['d', 'c'], continuous_dims=['X', 'Y'], categorical_dims='lg10_Z', additive=True)
    gp.build_model()
    assert isinstance(gp.model, pm.model.Model)
    assert all(name in gp.gp_dict.keys() for name in ['total', 'global', 'lg10_Z'])

# MAP estimation

def test_gp_fit_simple(example_gp):
    # Basically just makes sure that `fit` runs without errors
    gp = example_gp.fit(continuous_dims=['X', 'Y', 'lg10_Z'], continuous_levels={'lg10_Z': [8]})
    assert isinstance(gp.MAP, dict)


@pytest.mark.slow
def test_gp_fit(example_gp, fit_inputs, additive):
    # Basically just makes sure that `fit` runs without errors
    gp = example_gp.fit(**fit_inputs, additive=additive)
    assert isinstance(gp.MAP, dict)



"""Gaussian Process Model Building Interface"""

__version__ = '0.1.0'

from .regression import *
from .aggregation import *
from .arrays import *
from .plotting import *
from gumbi import style, utils, data

# Aliases
parray = ParameterArray
uarray = UncertainArray
uparray = UncertainParameterArray
mvuparray = MVUncertainParameterArray





from __future__ import annotations  # Necessary for self-type annotations until Python >3.10

import copy
import pickle
import warnings
import numpy as np
import pandas as pd

from scipy.stats import norm, lognorm, chi2, ncx2, rv_continuous, multivariate_normal
from scipy.stats._multivariate import multivariate_normal_frozen
from scipy.special import logit, expit

from uncertainties import unumpy as unp

from .utils import skip
from .aggregation import Standardizer

__all__ = ['LayeredArray', 'ParameterArray', 'UncertainArray', 'UncertainParameterArray', 'MVUncertainParameterArray']


class LogitNormal(rv_continuous):
    r"""A logit-normal continuous random variable.

    The probability density function for LogitNormal is:

    .. math::
        f \left( x, \sigma \right) = \frac{1}{\sigma \sqrt{2 \pi}} e^{- \frac{ \left( \text{logit} (x) - \mu \right)^2}{2 \sigma^2}} \frac{1}{x \left( 1-x \right)}

    for :math:`0<x<1`, :math:`\sigma>0`

    A logit-normal random variable `Y` can be parameterized in terms of the mean, :math:`\mu`, and standard deviation,
    :math:`\sigma`, of the unique normally distributed random variable `X` such that `expit(X) = Y`. This parametrization
    corresponds to setting `scale = σ` and `loc = expit(μ)`.
    """
    def __init__(self, loc=0.5, scale=1):
        super().__init__(self)
        self.scale = scale
        self.loc = logit(loc)

    def _pdf(self, x):
        return norm(loc=self.loc, scale=self.scale).pdf(logit(x)) / (x * (1 - x))

    def _cdf(self, x):
        return norm(loc=self.loc, scale=self.scale).cdf(logit(x))

    def ppf(self, q):
        return expit(norm(loc=self.loc, scale=self.scale).ppf(q))

    def rvs(self, size=None, random_state=None):
        return expit(norm(loc=self.loc, scale=self.scale).rvs(size=size, random_state=random_state))


class MultivariateNormalish(multivariate_normal_frozen):
    r"""A multivariate Normal distribution built from and callable on ParameterArrays.

    In particular, this class takes care of transforming variables to/from "natural" space as necessary.

    Parameters
    ----------
    mean : ParameterArray
        A ParameterArray containing the distribution mean, must be a single point
    cov : float or np.array, default 1
        Covariance matrix of the distribution as a standard array or scalar (default one)
    **kwargs
        Additional keyword arguments passed to `scipy.stats.multivariate_normal`
    """

    def __init__(self, mean: ParameterArray, cov: int | float | np.ndarray, **kwargs):
        assert isinstance(mean, ParameterArray), 'Mean must be a ParameterArray'
        if mean.ndim != 0:
            raise NotImplementedError('Multidimensional multivariate distributions are not yet supported.')

        self._names = mean.names
        self._stdzr = mean.stdzr
        self._log_vars = mean.stdzr.log_vars
        self._logit_vars = mean.stdzr.logit_vars
        self._islog = [1 if var in self._log_vars else 0 for var in self._names]
        self._islogit = [1 if var in self._logit_vars else 0 for var in self._names]

        super(MultivariateNormalish, self).__init__(mean=mean.z.values(), cov=cov, **kwargs)

    def pdf(self, x: ParameterArray) -> float:
        r"""Probability density function.

        Parameters
        ----------
        x : ParameterArray
            Quantiles, with the last axis of x denoting the components.

        Returns
        -------
        array-like
        """
        return super(MultivariateNormalish, self).pdf(x)

    def logpdf(self, x: ParameterArray) -> float:
        r"""Log of the probability density function.

        Parameters
        ----------
        x : ParameterArray
            Quantiles, with the last axis of x denoting the components.

        Returns
        -------
        array-like
        """
        return super(MultivariateNormalish, self).logpdf(x.z.dstack())

    def cdf(self, x: ParameterArray) -> float:
        r"""Cumulative distribution function.

        Parameters
        ----------
        x : ParameterArray
            Quantiles, with the last axis of x denoting the components.

        Returns
        -------
        array-like
        """
        return super(MultivariateNormalish, self).cdf(x.z.dstack())

    def logcdf(self, x: ParameterArray) -> float:
        r"""Log of the cumulative distribution function.

        Parameters
        ----------
        x : ParameterArray
            Quantiles, with the last axis of x denoting the components.

        Returns
        -------
        array-like
        """
        return super(MultivariateNormalish, self).logcdf(x)

    def rvs(self, size=1, random_state=None) -> ParameterArray:
        r"""Draw random samples from the distribution.

        Parameters
        ----------
        size: int or array-like of ints
        random_state: int

        Returns
        -------
        ParameterArray
        """
        # multivariate_normal throws a RuntimeWarning: covariance is not positive semidefinite if rvs() is called more
        # than once without calling dist().rvs() in between. Covariance isn't being altered as a side-effect of this
        # call, so this issue really makes no sense.....
        # with warnings.catch_warnings():
        #     warnings.filterwarnings("error")
        #     try:
        #         samples = super(MultivariateNormalish, self).rvs(size=size, random_state=random_state)
        #     except:
        #         multivariate_normal(mean=self.mean, cov=self.cov).rvs()
        #         samples = super(MultivariateNormalish, self).rvs(size=size, random_state=random_state)
        samples = super(MultivariateNormalish, self).rvs(size=size, random_state=random_state)
        return ParameterArray(**{p: samples[..., i] for i, p in enumerate(self._names)}, stdzd=True, stdzr=self._stdzr)


class LayeredArray(np.ndarray):
    """An array with one or more named values at every index.

    Parameters
    ----------
    name : str
    array : array-like
    """
    def __new__(cls, stdzr=None, **arrays):
        if arrays == {}:
            raise ValueError('Must supply at least one array')
        arrays = {name: np.asarray(array) for name, array in arrays.items() if array is not None}

        narray_dtype = np.dtype([(name, array.dtype) for name, array in arrays.items()])

        narray_prototype = np.empty(list(arrays.values())[0].shape, dtype=narray_dtype)
        for name, array in arrays.items():
            narray_prototype[name] = array

        larray = narray_prototype.view(cls)
        larray.names = list(narray_dtype.fields.keys())
        larray.stdzr = stdzr

        return larray

    def __array_finalize__(self, larray):
        if larray is None:
            return
        self.names = getattr(larray, 'names', None)
        self.stdzr = getattr(larray, 'stdzr', None)

    def __array_ufunc__(self, ufunc, method, *inputs, out=None, **kwargs):
        """https://numpy.org/doc/stable/user/basics.subclassing.html#array-ufunc-for-ufuncs"""
        args = []
        if len(set([larray.names[0] for larray in inputs if isinstance(larray, LayeredArray)]))>1:
            warnings.warnings.warn('Operating on arrays with different layer names, results may be unexpected.')
        for input_ in inputs:
            if isinstance(input_, LayeredArray):
                if len(input_.names) > 1:
                    raise ValueError('Cannot operate on array with multiple layer names')
                args.append(input_.astype(float).view(np.ndarray))
            else:
                args.append(input_)

        outputs = kwargs.get('out')
        if outputs:
            out_args = []
            for output in outputs:
                if isinstance(output, LayeredArray):
                    out_args.append(output.astype(float).view(np.ndarray))
                else:
                    out_args.append(output)
            kwargs['out'] = tuple(out_args)
        else:
            outputs = (None,) * ufunc.nout

        results = super().__array_ufunc__(ufunc, method, *args, **kwargs)

        if results is NotImplemented:
            return NotImplemented

        if ufunc.nout == 1:
            results = (results,)

        results = tuple(LayeredArray(**{self.names[0]: result})
                        if output is None else output
                        for result, output in zip(results,outputs))

        return results[0] if len(results) == 1 else results

    def __getitem__(self, item):
        default = super().__getitem__(item)
        if isinstance(item, str):
            arrays = {item: default}
        elif isinstance(item, int) or (isinstance(item, tuple) and all(isinstance(val, int) for val in item)):
            arrays = {name: value for name, value in zip(default.dtype.names, default)}
        elif isinstance(item, slice):
            arrays = {layer.names[0]: layer.values() for layer in default.as_list()}
        else:
            return default
        return LayeredArray(**arrays)

    def __repr__(self):
        return f'{tuple(self.names)}: {np.asarray(self)}'

    def get(self, name, default=None):
        """Return value given by `name` if it exists, otherwise return `default`."""
        if name in self.names:
            return self[name]
        elif default is None:
            return None
        else:
            return LayeredArray(**{name: default})

    def drop(self, name, missing_ok=True):
        if name in self.names:
            return LayeredArray(**{p: arr for p, arr in self.as_dict().items() if p != name})
        elif missing_ok:
            return self
        else:
            raise KeyError(f'Name {name} not found in array.')

    def values(self):
        """Values at each index stacked into regular ndarray."""
        stacked = np.stack([self[name].astype(float) for name in self.names])
        if len(self.names) > 1:
            return stacked
        else:
            return stacked[0]

    def dstack(self):
        """Values at each index as ndarrays stacked in sequence depth wise (along third axis)."""
        return np.dstack([la.values() for la in self.as_list()])

    def as_list(self, order=None):
        order = self.names if order is None else order
        assert all(name in order for name in self.names)
        return [self[name] for name in order]

    def as_dict(self):
        """Values corresponding to each named level as a dictionary."""
        return {name: self[name].values() for name in self.names}

    def add_layers(self, **arrays):
        """Add additional layers at each index."""
        arrays_ = arrays.as_dict() if isinstance(arrays, LayeredArray) else arrays
        return LayeredArray(**(self.as_dict() | arrays_))


class ParameterArray(LayeredArray):
    """Array of parameter values, allowing simple transformation.

    :class:`ParameterArray` stores not only the value of the variable itself but also a :class:`Standardizer` instance.
    This makes it simple to switch between the natural scale of the parameter and its transformed and standardized
    values through the :attr:`t` and :attr:`z` properties, respectively.

    This class can also be accessed through the alias :class:`parray`.

    Parameters
    ----------
    **arrays
        arrays to store with their names as keywords
    stdzr : Standardizer
        An instance  of :class:`Standardizer`
    stdzd : bool, default False
        Whether the supplied values are on standardized scale instead of the natural scale

    Examples
    --------

    A parray can created with a single parameter. In this case, `r` is treated as a `LogNormal` variable by the stdzr.

    >>> from gumbi import ParameterArray as parray
    >>> stdzr = Standardizer(d={'μ': -0.307, 'σ': 0.158}, log_vars=['d'])
    >>> rpa = parray(d=np.arange(5,10)/10, stdzr=stdzr)
    >>> rpa
    ('d',): [(0.5,) (0.6,) (0.7,) (0.8,) (0.9,)]
    >>> rpa.t
    ('r_t',): [(-0.69314718,) (-0.51082562,) (-0.35667494,) (-0.22314355,) (-0.10536052,)]
    >>> rpa.z
    ('r_z',): [(-2.4439695 ,) (-1.29003559,) (-0.31439838,) ( 0.53073702,) ( 1.27619927,)]

    If the parameter is completely absent from the stdzr, its natural, :attr:`t`, and :attr:`z` values are identical.

    >>> pa = parray(param=np.arange(5), stdzr=stdzr)
    >>> pa
    ('param',): [(0,) (1,) (2,) (3,) (4,)]
    >>> pa.t
    ('param_t',): [(0,) (1,) (2,) (3,) (4,)]
    >>> pa.z
    ('param_z',): [(0.,) (1.,) (2.,) (3.,) (4.,)]

    You can even do monstrous compositions like

    >>> np.min(np.sqrt(np.mean(np.square(rpa-rpa[0]-0.05)))).t
    ('r_t',): (-1.5791256,)

    If you `have` work with an ordinary numpy array, use :meth:`values`.

    >>> np.argmax(rpa.values())
    4

    Attributes
    ----------
    names : list of str
        Names of all stored parameters
    stdzr : Standardizer
        An instance  of :class:`Standardizer`
    """

    def __new__(cls, stdzr: Standardizer, stdzd=False, **arrays):
        if arrays == {}:
            raise ValueError('Must supply at least one array')

        if stdzd:
            arrays = {name: stdzr.unstdz(name, np.array(array))
                      for name, array in arrays.items()}

        parray = LayeredArray.__new__(cls, **arrays)
        parray.stdzr = stdzr

        return parray

    def __array_finalize__(self, parray):
        if parray is None:
            return
        self.stdzr = getattr(parray, 'stdzr', None)
        self.names = getattr(parray, 'names', None)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        result = super().__array_ufunc__(ufunc, method, *inputs, **kwargs)
        if result is NotImplemented:
            return NotImplemented
        return ParameterArray(**result.as_dict(), stdzr=self.stdzr, stdzd=False)

    def __getitem__(self, item):
        default = super(LayeredArray, self).__getitem__(item)
        if isinstance(item, str):
            arrays = {item: default}
        elif isinstance(item, int) or (isinstance(item, tuple) and all(isinstance(val, int) for val in item)):
            arrays = {name: value for name, value in zip(default.dtype.names, default)}
        elif isinstance(item, slice):
            arrays = {layer.names[0]: layer.values() for layer in default.as_list()}
        else:
            return default
        return ParameterArray(**arrays, stdzr=self.stdzr, stdzd=False)


    def get(self, name, default=None):
        """Return value given by `name` if it exists, otherwise return `default`"""
        if name in self.names:
            return self[name]
        elif default is None:
            return None
        else:
            return self.parray(**{name: default})

    def drop(self, name, missing_ok=True):
        if name in self.names:
            return self.parray(**{p: arr for p, arr in self.as_dict().items() if p != name})
        elif missing_ok:
            return self
        else:
            raise KeyError(f'Name {name} not found in array.')

    @property
    def z(self) -> LayeredArray:
        """Standardized values"""
        zdct = {name+'_z': self.stdzr.stdz(name, self[name].values()) for name in self.names}
        return LayeredArray(**zdct, stdzr=self.stdzr)

    @property
    def t(self) -> LayeredArray:
        """Transformed values"""
        tdct = {name+'_t': self.stdzr.transform(name, self[name].values()) for name in self.names}
        return LayeredArray(**tdct, stdzr=self.stdzr)

    def add_layers(self, stdzd=False, **arrays):
        """Add additional layers at each index"""
        narrays = super().add_layers(**arrays)
        if stdzd:
            for name in narrays.names:
                narrays[name] = self.stdzr.unstdz(name, narrays[name])
        return self.parray(**narrays.as_dict(), stdzd=False)

    def fill_with(self, **params):
        """Add a single value for a new parameter at all locations."""
        assert all([isinstance(value, (float, int)) for value in params.values()])
        assert all([isinstance(key, str) for key in params.keys()])
        return self.add_layers(**{param: np.full(self.shape, value) for param, value in params.items()})

    def parray(self, *args, **kwargs):
        """Create a new ParameterArray using this instance's standardizer"""
        return ParameterArray(*args, **kwargs, stdzr=self.stdzr)

    @classmethod
    def stack(cls, parray_list, axis=0, **kwargs):
        all_names = [pa.names for pa in parray_list]
        if not all(names == all_names[0] for names in all_names):
            raise ValueError('Arrays do not have the same names!')
        new = np.stack(parray_list, axis=axis, **kwargs)
        stdzr = parray_list[0].stdzr
        return cls(**{dim: new[dim] for dim in new.dtype.names}, stdzr=stdzr)

    @classmethod
    def vstack(cls, parray_list, **kwargs):
        all_names = [pa.names for pa in parray_list]
        if not all(names == all_names[0] for names in all_names):
            raise ValueError('Arrays do not have the same names!')
        new = np.vstack(parray_list, **kwargs)
        stdzr = parray_list[0].stdzr
        return cls(**{dim: new[dim] for dim in new.dtype.names}, stdzr=stdzr)

    @classmethod
    def hstack(cls, parray_list, **kwargs):
        all_names = [pa.names for pa in parray_list]
        if not all(names == all_names[0] for names in all_names):
            raise ValueError('Arrays do not have the same names!')
        new = np.hstack(parray_list, **kwargs)
        stdzr = parray_list[0].stdzr
        return cls(**{dim: new[dim] for dim in new.dtype.names}, stdzr=stdzr)


class UncertainArray(np.ndarray):
    """Structured array containing mean and variance of a normal distribution at each point.

    The main purpose of this object is to correctly `propagate uncertainty`_ under transformations. Arithmetic
    operations between distributions or between distributions and scalars are handled appropriately via the
    `uncertainties`_ package.

    Additionally, a `scipy Normal distribution`_ object can be created at each point through the :attr:`dist` property,
    allowing access to that objects such as :meth:`rvs`, :meth:`ppf`, :meth:`pdf`, etc.

    This class can also be accessed through the alias :class:`uarray`.

    .. _`propagate uncertainty`: https://en.wikipedia.org/wiki/Propagation_of_uncertainty
    .. _`uncertainties`: https://pythonhosted.org/uncertainties/
    .. _`scipy Normal distribution`: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.norm.html

    Notes
    -----
    The `name` argument is intended to be the general name of the value held, not unique to this instance. Combining two
    :class:`UncertainArray` objects with the same name results in a new object with that name; combining two objects
    with different names results in a new name that reflects this combination (so ``'A'+'B'`` becomes ``'(A+B)'``).

    Parameters
    ----------
    name : str
        Name of variable.
    μ : array
        Mean at each point
    σ2 : array
        Variance at each point
    **kwargs
        Names and values of additional arrays to store

    Examples
    --------
    >>> ua1 = UncertainArray('A', μ=1, σ2=0.1)
    >>> ua2 = uarray('A', μ=2, σ2=0.2)  #equivalent
    >>> ua2
    A['μ', 'σ2']: (2, 0.2)

    Addition of a scalar

    >>> ua1+1
    A['μ', 'σ2']: (2, 0.1)

    Addition and subtraction of two UncertainArrays:

    >>> ua2+ua1
    A['μ', 'σ2']: (3., 0.3)
    >>> ua2-ua1
    A['μ', 'σ2']: (1., 0.3)

    Note, however, that correlations are not properly accounted for (yet). Subtracting one UncertainArray from itself
    should give exactly zero with no uncertainty, but it doesn't:

    >>> ua2+ua2
    A['μ', 'σ2']: (4., 0.4)
    >>> ua2-ua2
    A['μ', 'σ2']: (0., 0.4)

    Mean of two `uarray` objects:

    >>> uarray.stack([ua1, ua2]).mean(axis=0)
    A['μ', 'σ2']: (1.5, 0.075)

    Mean within a single `uarray` object:

    >>> ua3 = uarray('B', np.arange(1,5)/10, np.arange(1,5)/100)
    >>> ua3
    B['μ', 'σ2']: [(0.1, 0.01) (0.2, 0.02) (0.3, 0.03) (0.4, 0.04)]
    >>> ua3.μ
    array([0.1, 0.2, 0.3, 0.4])
    >>> ua3.mean()
    B['μ', 'σ2']: (0.25, 0.00625)

    Adding two `uarrays` with differnt name creates an object with a new name

    >>> ua1+ua3.mean()
    (A+B)['μ', 'σ2']: (1.25, 0.10625)

    Accessing :attr:`dist` methods

    >>> ua3.dist.ppf(0.95)
    array([0.26448536, 0.43261743, 0.58489701, 0.72897073])
    >>> ua3.dist.rvs([3,*ua3.shape])
    array([[0.05361942, 0.14164882, 0.14924506, 0.03808633],
           [0.05804824, 0.09946732, 0.08727794, 0.28091272],
           [0.06291355, 0.47451576, 0.20756356, 0.2108717 ]])  # random

    Attributes
    ----------
    name : str
        Name of variable.
    fields : list of str
        Names of each level held in the array
    """

    def __new__(cls, name: str, μ: np.ndarray, σ2: np.ndarray, stdzr=None, **kwargs):
        μ_ = np.asarray(μ)
        σ2_ = np.asarray(σ2)
        assert(μ_.shape == σ2_.shape)
        base_dtypes = [('μ', μ_.dtype), ('σ2', σ2_.dtype)]
        extra_dtypes = [(dim, np.asarray(arr).dtype) for dim, arr in kwargs.items() if arr is not None]
        uarray_dtype = np.dtype(base_dtypes+extra_dtypes)

        uarray_prototype = np.empty(μ_.shape, dtype=uarray_dtype)
        uarray_prototype['μ'] = μ_
        uarray_prototype['σ2'] = σ2_
        for dim, arr in kwargs.items():
            if arr is not None:
                uarray_prototype[dim] = np.asarray(arr)

        uarray = uarray_prototype.view(cls)
        uarray.name = name
        uarray.stdzr = stdzr
        uarray.fields = list(uarray_dtype.fields.keys())

        return uarray

    def __array_finalize__(self, uarray):
        if uarray is None:
            return
        self.name = getattr(uarray, 'name', None)
        self.stdzr = getattr(uarray, 'stdzr', None)
        self.fields = getattr(uarray, 'fields', None)

    @property
    def μ(self) -> np.ndarray:
        """Nominal value (mean)"""
        return self['μ']

    @μ.setter
    def μ(self, val):
        self['μ'] = val

    @property
    def σ2(self) -> np.ndarray:
        """Variance"""
        return self['σ2']

    @σ2.setter
    def σ2(self, val):
        self['σ2'] = val

    @property
    def σ(self) -> np.ndarray:
        """Standard deviation"""
        return np.sqrt(self.σ2)

    @σ.setter
    def σ(self, val):
        self['σ2'] = val**2

    @property
    def _as_uncarray(self):
        return unp.uarray(self.μ, self.σ)

    @classmethod
    def _from_uncarray(cls, name, uncarray, **extra):
        return cls(name=name, μ=unp.nominal_values(uncarray), σ2=unp.std_devs(uncarray)**2, **extra)

    @property
    def dist(self) -> rv_continuous:
        """Array of :func:`scipy.stats.norm` objects"""
        return norm(loc=self.μ, scale=self.σ)

    @staticmethod
    def stack(uarray_list, axis=0) -> UncertainArray:
        new = np.stack(uarray_list, axis=axis)
        names = [ua.name for ua in uarray_list]
        if all(name == names[0] for name in names):
            name = uarray_list[0].name
        else:
            raise ValueError('Arrays do not have the same name!')
            # name = '('+', '.join(names)+')'
        return UncertainArray(name, **{dim: new[dim] for dim in new.dtype.names})

    def nlpd(self, target) -> float:
        """Negative log posterior density"""
        return -np.log(self.dist.pdf(target))

    def EI(self, target, best_yet, k=1) -> float:
        """Expected improvement

        Taken from https://github.com/akuhren/target_vector_estimation

        Parameters
        ----------
        target : float
        best_yet : float
        k : int

        Returns
        -------
        EI : float
        """

        nc = ((target - self.μ) ** 2) / self.σ2

        h1_nx = ncx2.cdf((best_yet / self.σ2), k, nc)
        h2_nx = ncx2.cdf((best_yet / self.σ2), (k+2), nc)
        h3_nx = ncx2.cdf((best_yet / self.σ2), (k+4), nc)

        t1 = best_yet * h1_nx
        t2 = self.σ2 * (k * h2_nx + nc * h3_nx)

        return t1 - t2

    def KLD(self, other):
        """Kullback–Leibler Divergence

        Parameters
        ----------
        other : UncertainArray

        Returns
        -------
        divergence : float
        """
        assert isinstance(other, UncertainArray)
        return np.log(other.σ / self.σ) + (self.σ2 + (self.μ - other.μ) ** 2) / (2 * other.σ2) - 1 / 2

    def BD(self, other):
        """Bhattacharyya Distance

        Parameters
        ----------
        other : UncertainArray

        Returns
        -------
        distance : float
        """
        assert isinstance(other, UncertainArray)
        return 1 / 4 * np.log(1 / 4 * (self.σ2 / other.σ2 + other.σ2 / self.σ2 + 2)) + 1 / 4 * ((self.μ - other.μ) ** 2 / (self.σ2 + other.σ2))


    def BC(self, other):
        """Bhattacharyya Coefficient

        Parameters
        ----------
        other : UncertainArray

        Returns
        -------
        coefficient : float
        """
        return np.exp(-self.BD(other))


    def HD(self, other):
        """Hellinger Distance

        Parameters
        ----------
        other : UncertainArray

        Returns
        -------
        distance : float
        """
        return np.sqrt(1 - self.BC(other))

    def __repr__(self):
        return f'{self.name}{self.fields}: {np.asarray(self)}'

    def __getitem__(self, item):
        default = super().__getitem__(item)
        if isinstance(item, int) or (isinstance(item, tuple) and all(isinstance(val, int) for val in item)):
            arrays = {name: value for name, value in zip(default.dtype.names, default)}
        elif isinstance(item, slice):
            # arrays = {layer.names[0]: layer.values() for layer in default.as_list()}
            return default
        else:
            return default.view(np.ndarray)
        return UncertainArray(self.name, **arrays)

    def sum(self, axis=None, dtype=None, out=None, keepdims=False, **kwargs) -> UncertainArray:
        """Summation with uncertainty propagation"""
        kwargs |= dict(axis=axis, dtype=dtype, out=out, keepdims=keepdims)
        new = self._as_uncarray.sum(**kwargs)
        extra = {dim: np.sum(self[dim]) for dim in self.fields if dim not in ['μ', 'σ2']}
        return self._from_uncarray(self.name, new, **extra)

    def mean(self, axis=None, dtype=None, out=None, keepdims=False, **kwargs) -> UncertainArray:
        """Mean with uncertainty propagation"""
        kwargs |= dict(axis=axis, dtype=dtype, out=out, keepdims=keepdims)
        new = self._as_uncarray.mean(**kwargs)
        extra = {dim: np.mean(self[dim]) for dim in self.fields if dim not in ['μ', 'σ2']}
        return self._from_uncarray(self.name, new, **extra)

    def __add__(self, other):
        new = self._as_uncarray
        if isinstance(other, UncertainArray):
            new += other._as_uncarray
            name = self.name if self.name == other.name else f'({self.name}+{other.name})'
        else:
            new += other
            name = self.name
        extra = {dim: np.mean(self[dim]) for dim in self.fields if dim not in ['μ', 'σ2']}
        return self._from_uncarray(name, new, **extra)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        new = self._as_uncarray
        if isinstance(other, UncertainArray):
            new -= other._as_uncarray
            name = self.name if self.name == other.name else f'({self.name}+{other.name})'
        else:
            new -= other
            name = self.name
        extra = {dim: np.mean(self[dim]) for dim in self.fields if dim not in ['μ', 'σ2']}
        return self._from_uncarray(name, new, **extra)

    def __rsub__(self, other):
        if isinstance(other, UncertainArray):
            new = other._as_uncarray
            name = self.name if self.name == other.name else f'({self.name}+{other.name})'
        else:
            new = copy.copy(other)
            name = self.name
        new -= self._as_uncarray
        extra = {dim: np.mean(self[dim]) for dim in self.fields if dim not in ['μ', 'σ2']}
        return self._from_uncarray(name, new, **extra)


class UncertainParameterArray(UncertainArray):
    r"""Structured array of parameter means and variances, allowing transformation with uncertainty handling.

    The primary role of this class is to compactly store the outputs of our regression models
    (e.g., :class:`gumbi.GP`). We typically use these models
    to produce parameter predictions or estimates, but under some transformation. For example, reaction `rate` must
    clearly be strictly positive, so we fit a GP to the `log` of rate in order to more appropriately conform to the
    assumption of normality. For prediction and visualization, however, we often need to switch back and forth between
    natural space (:math:`rate`), transformed space (:math:`\text{ln}\; rate`), and standardized space
    (:math:`\left( \text{ln}\; rate  - \mu_{\text{ln}\; rate} \right)/\sigma_{\text{ln}\; rate}`), meanwhile calculating
    summary statistics such as means and percentiles. This class is intended to facilitate switching between those
    different contexts.

    :class:`UncertainParameterArray`, also accessible through the alias :class:`uparray`, combines the functionality of
    :class:`ParameterArray` and :class:`UncertainArray`. A `uparray` stores the mean and variance of the
    variable itself as well as a :class:`Standardizer`
    instance. This  makes it simple to switch between the natural scale of the parameter and its transformed and
    standardized values through the :attr:`t` and :attr:`z` properties, respectively, with the accompanying variance
    transformed and scaled appropriately. This uncertainty is propagated under transformation, as with
    :class:`UncertainArray`, and a scipy distribution object can be created at each point through the :attr:`dist`
    property, allowing access to that objects such as :meth:`rvs`, :meth:`ppf`, :meth:`pdf`, etc.

    Notes
    -----
    The `name` argument is intended to be the general name of the value held, not unique to this instance. Combining two
    :class:`UncertainParameterArray` objects with the same name results in a new object with that name; combining two
    objects with different names results in a new name that reflects this combination (so ``'A'+'B'`` becomes
    ``'(A+B)'``).

    The behavior of this object depends on the transformation associated with it, as indicated by its `name` in its
    stored :class:`Standardizer` instance. If this transformation is :func:`np.log`, the parameter is treated as a
    `LogNormal` variable; otherwise it's treated as a `Normal` variable. This affects which distribution is returned by
    :attr:`dist` (`lognorm`_ vs `norm`_) and also the interpretation of :attr:`μ` and :attr:`σ2`.

     * For a `Normal` random
       variable, these are simply parameter's mean and variance in unstandardized space, :attr:`t.μ` and :attr:`t.σ2`
       are identical to :attr:`μ` and :attr:`σ2`, and :attr:`z.μ` and :attr:`z.σ2` are the parameter's mean and variance
       in standardized space.
     * For a `LogNormal` random variable ``Y``, however, :attr:`t.μ` and :attr:`t.σ2` are the mean and variance of a
       `Normal` variable ``X`` such that ``exp(X)=Y`` (:attr:`z.μ` and :attr:`z.σ2` are this mean and variance in
       standardized space). In this case, :attr:`μ` and :attr:`σ2` are the scale and shape descriptors of ``Y``, so
       ``self.μ = np.exp(self.t.μ)`` and ``self.σ2 = self.t.σ2``. Thus, :attr:`μ` and :attr:`σ2` are not strictly the
       mean and variance of the random variable in natural space, these can be obtained from the :attr:`dist`.

       * This behavior is most important, and potentially most confusing, when calculating the :meth:`mean`. Averaging
         is performed in `transformed` space, where the random variable exhibits a `Normal` distribution and the mean
         also exhibits a `Normal` distribution, allowing error propagation to be applied analytically. The :attr:`μ` and
         :attr:`σ2` returned are the descriptors of the `LogNormal` distribution that represents the reverse
         transformation of this new `Normal` distribution. Therefore, the result is more akin to marginalizing out the
         given dimensions in the underlying model than a true natural-space average.

    .. _norm: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.norm.html
    .. _lognorm: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.lognorm.html

    See Also
    --------
    `norm <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.norm.html>`_: \
    scipy `Normal` random variable

    `lognorm <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.lognorm.html>`_: \
    scipy `LogNormal` random variable

    :class:`ParameterArray`

    :class:`UncertainArray`

    :class:`Standardizer`

    Parameters
    ----------
    name : str
        Name of variable.
    μ : array
        Mean at each point
    σ2 : array
        Variance at each point
    stdzr : Standardizer
        An instance  of :class:`Standardizer`, converted internally to :class:`Standardizer`
    stdzd : bool, default False
        Whether the supplied values are on standardized scale instead of the natural scale

    Examples
    --------
    Create a `LogNormal` random variable, as indicated by its :class:`Standardizer`

    >>> from gumbi import uparray, Standardizer
    >>> import numpy as np
    >>> stdzr = Standardizer(m = {'μ': -5.30, 'σ': 0.582}, log_vars=['c'])
    >>> upa = uparray('c', np.arange(1,5)/10, np.arange(1,5)/100, stdzr)
    >>> upa
    m['μ', 'σ2']: [(0.1, 0.01) (0.2, 0.02) (0.3, 0.03) (0.4, 0.04)]
    >>> stdzr.transforms['c']
    [<ufunc 'log'>, <ufunc 'exp'>]

    Mean and variance of the parameter in standardized space:

    >>> upa.z
    m_z['μ', 'σ2']: [(5.15019743, 0.02952256) (6.34117197, 0.05904512)
                     (7.03784742, 0.08856768) (7.53214651, 0.11809024)]

    Verify round-trip transformation:

    >>> upa.stdzr.unstdz(upa.name, upa.z.μ, upa.z.σ2)
    (array([0.1, 0.2, 0.3, 0.4]), array([0.01, 0.02, 0.03, 0.04]))

    Create a `uparray` from already-standardized values and verify round-trip transformation:

    >>> uparray('c', np.arange(-2,3), np.arange(1,6)/10, stdzr, stdzd=True).z
    m_z['μ', 'σ2']: [(-2., 0.1) (-1., 0.2) ( 0., 0.3) ( 1., 0.4) ( 2., 0.5)]

    For `LogNormal` parameters, uparray follows the `scipy.stats` convention  of parameterizing a lognormal random
    variable in terms of it's natural-space mean and its log-space standard deviation. Thus, a LogNormal uparray defined
    as `m['μ', 'σ2']: (0.1, 0.01)` represents `exp(Normal(log(0.1), 0.01))`.

    Note that the mean is not simply the mean of each component, it is the parameters of the `LogNormal` distribution
    that corresponds to the mean of the underlying `Normal` distributions in `log` (transformed) space.

    >>> upa.μ.mean()
    0.25
    >>> upa.σ2.mean()
    0.025
    >>> upa.mean()
    m['μ', 'σ2']: (0.22133638, 0.00625)

    You can verify the mean and variance returned by averaging over the random variable explicitly.

    >>> upa.mean().dist.mean()
    2.2202914201059437e-01
    >>> np.exp(upa.t.mean().dist.rvs(10000, random_state=2021).mean())
    2.2133371283050837e-01
    >>> upa.mean().dist.var()
    3.0907071428047016e-04
    >>> np.log(upa.mean().dist.rvs(10000, random_state=2021)).var()
    6.304628046829242e-03

    Calculate percentiles

    >>> upa.dist.ppf(0.025)
    array([0.08220152, 0.1515835 , 0.21364308, 0.27028359])
    >>> upa.dist.ppf(0.975)
    array([0.12165225, 0.26388097, 0.42126336, 0.59197082])

    Draw samples

    >>> upa.dist.rvs([3, *upa.shape], random_state=2021)
    array([[0.11605116, 0.22006429, 0.27902589, 0.34041327],
           [0.10571616, 0.1810085 , 0.36491077, 0.45507622],
           [0.10106982, 0.21230397, 0.3065239 , 0.33827997]])

    You can compose the variable with numpy functions, though you may get a warning if the operation is poorly defined
    for the distribution (which is most transforms on `LogNormal` distributions). Transformations are applied in
    transformed space.

    >>> (upa+1+np.tile(upa, (3,1))[2,3]).mean().t.dist.ppf(0.5)
    UserWarning: Transform is poorly defined for <ufunc 'log'>; results may be unexpected.
    -1.8423623672812148


    Attributes
    ----------
    name : str
        Name of variable.
    μ : array
        Mean at each point
    σ2 : array
        Variance at each point
    fields : list of str
        Names of each level held in the array
    stdzr : Standardizer
        An instance  of :class:`Standardizer` created from the supplied :class:`Standardizer` object
    """

    def __new__(cls, name: str, μ: np.ndarray, σ2: np.ndarray, stdzr: Standardizer, stdzd=False):
        μ_ = np.asarray(μ)
        σ2_ = np.asarray(σ2)
        assert(μ_.shape == σ2_.shape)

        if stdzd:
            μ_, σ2_ = stdzr.unstdz(name, μ_, σ2_)

        uparray_dtype = np.dtype([('μ', μ_.dtype), ('σ2', σ2_.dtype)])

        uparray_prototype = np.empty(μ_.shape, dtype=uparray_dtype)
        uparray_prototype['μ'] = μ_
        uparray_prototype['σ2'] = σ2_

        uparray = uparray_prototype.view(cls)
        uparray.name = name
        uparray.stdzr = stdzr
        uparray.fields = list(uparray_dtype.fields.keys())

        return uparray

    def __array_finalize__(self, uparray):
        if uparray is None:
            return
        self.name = getattr(uparray, 'name', None)
        self.fields = getattr(uparray, 'fields', None)
        self.stdzr = getattr(uparray, 'stdzr', None)

    @property
    def z(self) -> UncertainArray:
        """Standardized values"""

        zmean, zvar = self.stdzr.stdz(self.name, self.μ, self.σ2)

        return UncertainArray(f'{self.name}_z', zmean, zvar, stdzr=self.stdzr)

    @property
    def t(self) -> UncertainArray:
        """Transformed values"""

        tmean, tvar = self.stdzr.transform(self.name, self.μ, self.σ2)

        return UncertainArray(f'{self.name}_t', tmean, tvar, stdzr=self.stdzr)

    @property
    def _ftransform(self):
        return self.stdzr.transforms.get(self.name, [skip, skip])[0]

    @property
    def _as_uncarray(self):
        return unp.uarray(self.z.μ, self.z.σ)

    def _from_uncarray(self, name, uncarray):
        z = UncertainArray._from_uncarray(name, uncarray)
        return self._from_z(z)

    @property
    def dist(self) -> rv_continuous:
        """Array of :func:`scipy.stats.rv_continuous` objects.

        If the transformation associated with the array's parameter is log/exp, this is a `lognorm` distribution object
        with ``scale=self.μ`` and ``s=self.t.σ``. Otherwise it is a `norm` distribution with ``loc=self.μ`` and
        ``scale=self.σ``. See the scipy documentation on `LogNormal`_ and `Normal`_ random variables for more
        explanation and a list of methods.

        .. _Normal: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.norm.html
        .. _LogNormal: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.lognorm.html
        """
        dists = {
            skip: super().dist,
            np.log: lognorm(scale=self.μ, s=self.σ),
            logit: LogitNormal(loc=self.μ, scale=self.σ)
        }
        return dists[self._ftransform]

    def sum(self, axis=None, dtype=None, out=None, keepdims=False, **kwargs):
        self._warn_if_poorly_defined()
        kwargs |= dict(axis=axis, dtype=dtype, out=out, keepdims=keepdims)
        z = self.z.sum(**kwargs)
        return self._from_z(z)

    def mean(self, axis=None, dtype=None, out=None, keepdims=False, **kwargs):
        """The natural-space distribution parameters which represent the mean of the transformed-space distributions"""
        kwargs |= dict(axis=axis, dtype=dtype, out=out, keepdims=keepdims)
        z = self.z.mean(**kwargs)
        return self._from_z(z)

    def _from_z(self, z):
        name = z.name.replace('_z', '')
        fields = {dim: z[dim] for dim in z.fields}
        return UncertainParameterArray(name, **fields, stdzr=self.stdzr, stdzd=True)

    def _from_t(self, t):
        name = t.name.replace('_t', '')
        t.μ, t.σ2 = self.stdzr.untransform(name, t.μ, t.σ2)
        return UncertainParameterArray(name, **{dim: t[dim] for dim in t.fields}, stdzr=self.stdzr,
                                       stdzd=False)

    def _warn_if_dissimilar(self, other):
        if isinstance(other, UncertainParameterArray):
            if not self.stdzr == other.stdzr:
                warnings.warn('uparrays have dissimilar Standardizers')

    def _warn_if_poorly_defined(self):
        if self._ftransform is not skip:
            warnings.warn(f'Transform is poorly defined for {self._ftransform}; results may be unexpected.')

    # def __repr__(self):
    #     return f'{self.name}{self.fields}: {np.asarray(self)}'

    def __getitem__(self, item):
        default = super(UncertainArray, self).__getitem__(item)
        if isinstance(item, int) or (isinstance(item, tuple) and all(isinstance(val, int) for val in item)):
            arrays = {name: value for name, value in zip(default.dtype.names, default)}
        elif isinstance(item, slice):
            # arrays = {layer.names[0]: layer.values() for layer in default.as_list()}
            return default
        else:
            return default.view(np.ndarray)
        return UncertainParameterArray(self.name, stdzr=self.stdzr, stdzd=False, **arrays)

    def __add__(self, other):
        self._warn_if_dissimilar(other)
        self._warn_if_poorly_defined()
        if isinstance(other, UncertainParameterArray):
            new = self._from_t(self.t.__add__(other.t))
            new.stdzr = Standardizer(**(self.stdzr | other.stdzr))
        else:
            new = super().__add__(other)
        return new

    def __sub__(self, other):
        self._warn_if_dissimilar(other)
        self._warn_if_poorly_defined()
        if isinstance(other, UncertainParameterArray):
            new = self._from_t(self.t.__sub__(other.t))
            new.stdzr = Standardizer(**(self.stdzr | other.stdzr))
        else:
            new = super().__sub__(other)
        return new

    def __rsub__(self, other):
        self._warn_if_dissimilar(other)
        self._warn_if_poorly_defined()
        if isinstance(other, UncertainParameterArray):
            new = self._from_t(self.t.__rsub__(other.t))
            new.stdzr = Standardizer(**(other.stdzr | self.stdzr))
        else:
            new = super().__rsub__(other)
        return new


class MVUncertainParameterArray(np.ndarray):
    r"""Structured array of multiple parameter means and variances along with correlations.

    This class is essentially a combination of the :class:`ParameterArray` and :class:`UncertainParameterArray` classes.

    See Also
    --------
    `multivariate_normal <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.multivariate_normal.html>`_:
    scipy `Multivariate Normal` random variable

    :class:`ParameterArray`

    :class:`UncertainParameterArray`

    Parameters
    ----------
    \*uparrays: UncertainParameterArray
        UParrays of identical shape containing the marginal mean and variance of each parameter
    stdzr : Standardizer
        An instance  of :class:`Standardizer`
    stdzd : bool, default False
        Whether the supplied values are on standardized scale instead of the natural scale

    Examples
    --------
    Create an MVUParray from two UParrays with negative correlation

    >>> import numpy as np
    >>> from gumbi import uparray, mvuparray, Standardizer
    >>>
    >>> stdzr = Standardizer(m = {'μ': -5.30, 'σ': 0.582}, r = {'μ': -0.307, 'σ': 0.158}, log_vars=['d', 'c'])
    >>> m_upa = uparray('c', np.arange(1,5)/10, np.arange(1,5)/100, stdzr)
    >>> r_upa = uparray('d', np.arange(1,5)/10+0.5, np.arange(1,5)/100*2, stdzr)
    >>> cor = np.array([[1, -0.6], [-0.6, 1]])
    >>> cor
    array([[ 1. , -0.6],
           [-0.6,  1. ]])
    >>> mvup = mvuparray(m_upa, r_upa, cor=cor)

    The MVUParray is displayed as an array of tuples ((μ_a, μ_b), (σ2_a, σ2_b))

    >>> mvup
    ('c', 'd')['μ', 'σ2']: [((0.1, 0.6), (0.01, 0.02)) ((0.2, 0.7), (0.02, 0.04))
                            ((0.3, 0.8), (0.03, 0.06)) ((0.4, 0.9), (0.04, 0.08))]

    The marginal means or variances can be extracted as ParameterArrays:

    >>> mvup.μ
    ('c', 'd'): [(0.1, 0.6) (0.2, 0.7) (0.3, 0.8) (0.4, 0.9)]

    The component UncertainParameterArrays can be extracted with the :meth:`get` method:

    >>> mvup.get('d')
    r['μ', 'σ2']: [(0.6, 0.02) (0.7, 0.04) (0.8, 0.06) (0.9, 0.08)]

    Slicing and indexing works as normal:

    >>> mvup[::2]
    ('c', 'd'): [((0.1, 0.6), (0.01, 0.02)) ((0.3, 0.8), (0.03, 0.06))]

    Transformed and standardized distributions can be obtained as with UncertainParameterArray

    >>> mvup.t
    ('m_t', 'r_t')['μ', 'σ2']: [((-2.30258509, -0.51082562), (0.01, 0.02))
                                ((-1.60943791, -0.35667494), (0.02, 0.04))
                                ((-1.2039728 , -0.22314355), (0.03, 0.06))
                                ((-0.91629073, -0.10536052), (0.04, 0.08))]
    >>> mvup.z
    ('m_z', 'r_z')['μ', 'σ2']: [((5.15019743, -1.29003559), (0.02952256, 0.80115366))
                                ((6.34117197, -0.31439838), (0.05904512, 1.60230732))
                                ((7.03784742,  0.53073702), (0.08856768, 2.40346098))
                                ((7.53214651,  1.27619927), (0.11809024, 3.20461465))]

    For 0-d MVUParrays (or individual elements of larger arrays), the :meth:`dist` exposes the scipy
    `multivariate_normal` object. Because this distribution is defined in standardized space (by default, `stdzd=True`)
    or transformed spaced (`stdzd=False`), ParameterArrays may be the most convenient way to pass arguments to the
    distribution methods:

    >>> pa = mvup.parray(m=0.09, d=0.61)
    >>> mvup[0].dist().cdf(pa.z.values())
    0.023900979112885523
    >>> mvup[0].dist(stdzd=False).cdf(pa.t.values())
    0.023900979112885523

    Perhaps most importantly, the :meth:`rvs` allows drawing correlated samples from the joint distribution, returned as
    a ParameterArray:

    >>> mvup[0].dist.rvs(10, random_state=2021)
    ('c', 'd'): [(0.0962634 , 0.74183363) (0.09627651, 0.56437764)
                 (0.09140986, 0.64790721) (0.09816149, 0.70518567)
                 (0.10271404, 0.60974628) (0.09288982, 0.60933939)
                 (0.0983131 , 0.63588871) (0.12262933, 0.45941758)
                 (0.1070759 , 0.4918009 ) (0.11118635, 0.49708401)]
    >>> mvup[0].dist.rvs(1, random_state=2021).z
    ('m_z', 'r_z'): (5.08476443, 0.05297293)

    Allowing us to easily visualize the joint distribution:

    >>> import pandas as pd
    >>> import seaborn as sns
    >>> sns.jointplot(x='c', y='d', data=pd.DataFrame(mvup[0].dist.rvs(1000).as_dict()), kind='kde')

    """

    def __new__(cls, *uparrays, cor, stdzr=None):

        shape = uparrays[0].shape
        assert all([upa.shape == shape for upa in uparrays])
        assert cor.shape[0] == len(uparrays)
        stdzr = uparrays[0].stdzr if stdzr is None else stdzr

        μ_ = ParameterArray(**{upa.name: upa.μ for upa in uparrays}, stdzr=stdzr)
        σ2_ = ParameterArray(**{upa.name: upa.σ2 for upa in uparrays}, stdzr=stdzr)

        mvuparray_dtype = np.dtype([('μ', μ_.dtype), ('σ2', σ2_.dtype)])
        mvuparray_prototype = np.empty(shape, dtype=mvuparray_dtype)
        mvuparray_prototype['μ'] = μ_
        mvuparray_prototype['σ2'] = σ2_

        mvuparray = mvuparray_prototype.view(cls)
        mvuparray.names = [upa.name for upa in uparrays]
        mvuparray.stdzr = stdzr
        mvuparray.fields = list(mvuparray_dtype.fields.keys())
        mvuparray.cor = cor

        return mvuparray

    def __array_finalize__(self, mvup):
        if mvup is None:
            return
        self.names = getattr(mvup, 'names', None)
        self.fields = getattr(mvup, 'fields', None)
        self.stdzr = getattr(mvup, 'stdzr', None)
        self.cor = getattr(mvup, 'cor', None)

    def __repr__(self):
        return f'{tuple(self.names)}{self.fields}: {np.asarray(self)}'

    def __getitem__(self, item):
        default = super().__getitem__(item)
        if isinstance(item, int) or (isinstance(item, tuple) and all(isinstance(val, int) for val in item)):
            arrays = [self.get(name)[item] for name in self.names]
        elif isinstance(item, slice):
            # arrays = [self.get(name)[item] for name in self.names]
            # Not sure why this "just works" for slices...
            return default
        else:
            return default.view(ParameterArray)
        return self.mvuparray(*arrays)

    def get(self, name, default=None) -> UncertainParameterArray | MVUncertainParameterArray:
        """Return one component parameter as an UncertainParameterArray or a subset as an MVUncertainParameterArray"""
        if isinstance(name, str):
            if name in self.names:
                return self.uparray(name, self['μ'][name].values(), self['σ2'][name].values())
            else:
                return default
        elif isinstance(name, list):
            idxs = [self.names.index(n) for n in name]
            return self.mvuparray([self.get(n) for n in name], cor=self.cor[idxs, :][:, idxs])

    @property
    def μ(self) -> ParameterArray:
        """Means"""
        return self['μ']

    @μ.setter
    def μ(self, val):
        self['μ'] = val

    @property
    def σ2(self) -> ParameterArray:
        """Marginal variances"""
        return self['σ2']

    @σ2.setter
    def σ2(self, val):
        self['σ2'] = val

    @property
    def σ(self) -> ParameterArray:
        """Standard deviations"""
        return self.parray(**{k: np.sqrt(v) for k, v in self['σ2'].as_dict().items()})

    @property
    def t(self) -> MVUncertainParameterArray:
        """Transformed values.

        Returns a MVUncertainParameterArray with the same Standardizer values as the current instance but with all
        transforms set to `lambda x: x`.
        """
        stdzr = Standardizer(**{k+'_t': v for k, v in self.stdzr.items()})
        return self.mvuparray(*[self.get(name).t for name in self.names], stdzr=stdzr)

    @property
    def z(self) -> MVUncertainParameterArray:
        """Standardized values.

        Returns a MVUncertainParameterArray with a default Standardizer (mean and SD of all variables set to zero and
        all transforms set to `lambda x: x`).
        """
        stdzr = Standardizer(**{k+'_z': {'μ': 0, 'σ2': 1} for k in self.names})
        return self.mvuparray(*[self.get(name).z for name in self.names], stdzr=stdzr)

    def parray(self, *args, **kwargs) -> ParameterArray:
        """Create a ParameterArray using this instance's Standardizer"""
        kwargs.setdefault('stdzr', self.stdzr)
        return ParameterArray(*args, **kwargs)

    def uparray(self, *args, **kwargs) -> UncertainParameterArray:
        """Create an UncertainParameterArray using this instance's Standardizer"""
        kwargs.setdefault('stdzr', self.stdzr)
        return UncertainParameterArray(*args, **kwargs)

    def mvuparray(self, *args, **kwargs) -> MVUncertainParameterArray:
        """Create an MVUncertainParameterArray using this instance's Standardizer"""
        kwargs.setdefault('stdzr', self.stdzr)
        kwargs.setdefault('cor', self.cor)
        return MVUncertainParameterArray(*args, **kwargs)

    def cov(self, stdzd=True, whiten=1e-10):
        """Covariance matrix (only supported for 0-D MVUParrays)"""
        # TODO: numpy versions > 1.19.3 can have bizarre inscrutable errors when handling mvup.cov. Monitor for fixes.
        if np.__version__ > '1.19.3':
            warnings.warn('numpy version >1.19.3 may lead to inscrutable linear algebra errors with mvup.cov. May just be on Windows/WSL. Hopefully fixed soon.')
        if self.ndim != 0:
            raise NotImplementedError('Multidimensional multivariate covariance calculations are not yet supported.')

        σ = self.z.σ.values() if stdzd else self.t.σ.values()

        cov = np.diag(σ) @ self.cor @ np.diag(σ)

        if whiten:
            cov += whiten*np.eye(*cov.shape)

        return cov

    @property
    def dist(self) -> MultivariateNormalish:
        """Scipy :func:`multivariate_normal` object (only supported for 0-D MVUParrays)"""
        # TODO: numpy versions > 1.19.3 can have bizarre inscrutable errors when handling mvup.dist. Monitor for fixes.
        if np.__version__ > '1.19.3':
            warnings.warn('numpy version >1.19.3 may lead to inscrutable linear algebra errors with mvup.dist. May just be on Windows/WSL. Hopefully fixed soon.')
        if self.ndim != 0:
            raise NotImplementedError('Multidimensional multivariate distributions are not yet supported.')
        return MultivariateNormalish(mean=self.μ, cov=self.cov(stdzd=True))

    def mahalanobis(self, parray: ParameterArray) -> float:
        """Calculates the Mahalanobis distance between the MVUParray distribution and a (ParameterArray) point."""
        cov_inv = np.linalg.inv(self.cov(stdzd=True))
        points = np.stack([parray.z.get(p+'_z').values() for p in self.names])
        μ = np.stack([self.z.μ.get(p+'_z').values() for p in self.names])
        diff = points-μ
        return np.sqrt(diff.T @ cov_inv @ diff)

    def outlier_pval(self, parray: ParameterArray) -> float:
        """Calculates the p-value that a given (ParameterArray) point is an outlier from the MVUParray distribution."""
        MD = self.mahalanobis(parray)
        n_params = len(self.names)
        pval = 1-chi2.cdf(MD**2, df=n_params)
        return pval




"""Utility functions for working with pymc3"""

import pymc3 as pm


def nc_Normal(name: str, mu: float, sigma: float, **kws):
    """Constructs non-centered normal distribution via :math:`\mathcal{N}(0,1)\cdot\sigma+\mu`

    Parameters
    ----------
    name: str
        Name given to pymc3 RV.
    mu: float
        Mean.
    sigma: float
        Standard deviation
    **kws:
        Additional arguments passed to pm.Normal.

    Returns
    -------
    rv: pm.Deterministic
        Non-centered Normal distribution
    nc: pm.Normal
        Underlying standard Normal distribution, named with '_nc' appended

    """
    nc = pm.Normal(name+'_nc', mu=0, sigma=1, **kws)
    rv = pm.Deterministic(name, mu + sigma*nc)

    return rv, nc


def sc_Exponential(name: str, mu: float, **kws):
    """Scaled Exponential distribution: :math:`\text{Exponential}(1)\cdot\mu

    Parameters
    ----------
    name: str
        Name given to pymc3 RV.
    mu: float
        Mean.
    **kws:
        Additional arguments passed to pm.Exponential.

    Returns
    -------
    rv: pm.Deterministic
        Scaled Exponential distribution
    nc: pm.Exponential
        Underlying standard Exponential distribution, named with '_nc' appended

    """
    nc = pm.Exponential(name+'_nc', lam=1, **kws)
    rv = pm.Deterministic(name, mu*nc)
    return rv, nc



import numpy as np
from scipy.spatial.distance import pdist
from scipy.stats import ncx2


def get_ℓ_prior(points):
    distances = pdist(points[:, None])
    distinct = distances != 0
    ℓ_l = distances[distinct].min() if sum(distinct) > 0 else 0.1
    ℓ_u = distances[distinct].max() if sum(distinct) > 0 else 1
    ℓ_σ = max(0.1, (ℓ_u - ℓ_l) / 6)
    ℓ_μ = ℓ_l + 3 * ℓ_σ
    return ℓ_μ, ℓ_σ


"""Generic utility functions"""

import os
import pathlib as pl
import matplotlib.pyplot as plt


def setup_paths(make_missing=True):
    """Gathers paths of expected directories in experiment structure

    Parameters
    ----------
    make_missing: bool, default True
        Whether to create directories that are expected but not found

    Returns
    -------
    dict
        Pathlib objects for each directory: 'Base' (root directory for experiment), 'Code', 'Data', 'Results', and
        'Figures'
    """
    code_pth = pl.Path(os.getcwd())
    base_pth = code_pth.parent
    data_pth = base_pth / 'Data'
    rslt_pth = base_pth / 'Results'
    fig_pth = base_pth / 'Figures'
    if make_missing:
        data_pth.mkdir(parents=True, exist_ok=True)
        rslt_pth.mkdir(parents=True, exist_ok=True)
        fig_pth.mkdir(parents=True, exist_ok=True)
    return base_pth, code_pth, data_pth, rslt_pth, fig_pth


def savefig(filename: str, fig=None, path=None, silent=False, **kwargs):
    """Saves figure as .png (300 dpi, transparent) and .svg

    Parameters
    ----------
    filename : str
        Name of file to save (without extension)
    fig : mpl.figures.Figure, optional
        Figure object to save. If `None`, calls `plt.gcf()`
    path : pl.Path, optional
        Directory in which to save figure
    silent : bool, default False
        Whether to display figure saving progress
    **kwargs
        Additional arguments passed to `mpl.figures.Figure.savefig()`
    """

    fig = plt.gcf() if fig is None else fig
    path = setup_paths(make_missing=False)[-1] if path is None else path

    kwargs.setdefault('bbox_inches', 'tight')
    kwargs.setdefault('transparent', True)

    if not silent:
        print('Saving.', end='')
    fig.savefig(path / (filename + '.png'), dpi=300, **kwargs)
    if not silent:
        print('.', end='')
    fig.savefig(path / (filename + '.svg'), **kwargs)
    if not silent:
        print('Done')



"""Internal utility functions, not intended for public consumption"""
from functools import wraps
import numpy as np

from typing import Iterable


def skip(x):
    """Returns input unmodified"""
    return x


# Round to a specified number of significant digits
def round_to_n(x, n=2):
    if isinstance(x, float):
        rounded = np.round(x, -int(np.floor(np.log10(np.abs(x)))-(n-1)))
    elif isinstance(x, (list, np.ndarray)):
        rounded = [np.round(x_, -int(np.floor(np.log10(np.abs(x_)))-(n-1))) if x_ != 0 else 0 for x_ in x]
    else:
        raise ValueError('x must be float, list, or ndarray.')
    return np.where(x == 0., x, rounded)


def NotImplementedWrapper(func):
    """Blocks execution and raises an error"""
    @wraps(func)
    def block(*args, **kwargs):
        """Raises error"""
        raise NotImplementedError
    return block


def assert_in(name: str, arg, lst: Iterable):
    """Raises error if value not in list"""
    if arg not in lst:
        raise ValueError(f'{name} must be one of {lst}')


def assert_is_subset(name: str, subset: Iterable, superset: Iterable):
    """Raises error if any required value not in list"""
    l_set = set(superset)
    r_set = set(subset)
    if not l_set.issuperset(r_set):
        missing = list(r_set.difference(l_set))
        msg = f'{list_is_are(missing)} missing from {name}'
        raise ValueError(msg)


def assert_one(names: str, lst: Iterable):
    """Raises error if none or multiple of values in `lst` are not None"""
    if sum(el is not None for el in lst) != 1:
        raise ValueError(f'Exactly one of {names} must be supplied')


def list_is_are(lst: list) -> str or None:
    """String formats list to be grammatically correct"""
    if len(lst) == 0:
        return None
    elif len(lst) == 1:
        msg = f'{list_and(lst)} is'
    else:
        msg = f'{list_and(lst)} are'
    return msg


def list_and(lst: list) -> str:
    if len(lst) == 1:
        msg = f'{lst[0]}'
    else:
        msg = f'{", ".join(lst[:-1])} and {lst[-1]}'
    return msg


class Trigger(object):
    """Descriptor to create a property that calls an instance method when its value is changed.
    Note, must be specified as a class variable.

    Parameters
    ----------
    method: str
        The name of the instance method to be called
    default: optional
        Default value of variable
    """

    def __init__(self, method, default=None):
        self.default = default
        self.method = method
        self.data = dict()

    def __get__(self, instance, owner):
        return self.data.get(instance, self.default)

    def __set__(self, instance, value):
        self.data[instance] = value
        getattr(instance, self.method)()


class InstanceCopy:
    """A class that creates instances by copying all attributes of a parent instance"""

    def __init__(self, parent):
        assert isinstance(parent, self.__class__.__bases__[-1])

        for attr in parent.__dict__:
            setattr(self, attr, getattr(parent, attr))



from __future__ import annotations  # Necessary for self-type annotations until Python >3.10

import pickle
import warnings
from collections import namedtuple
from dataclasses import dataclass
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
from scipy.special import logit, expit

from .utils import skip

__all__ = ['Standardizer', 'TidyData', 'WideData', 'DataSet']


class Standardizer(dict):
    r"""Container for dict of mean (μ) and variance (σ2) for every parameter.

    :class:`Standardizer` objects allow transformation and normalization of datasets. The main methods are :meth:`stdz`,
    which attempts to coerce the values of a given variable to a standard normal distribution (`z-scores`), and its
    complement :meth:`unstdz`. The steps are

    .. math::
        \mathbf{\text{tidy}} \rightarrow \text{transform} \rightarrow \text{mean-center} \rightarrow \text{scale}
        \rightarrow \mathbf{\text{tidy.z}}

    For example, reaction `rate` must clearly be strictly positive, so we use a `log` transformation so that it behaves
    as a normally-distributed random variable. We then mean-center and scale this transformed value to obtain `z-scores`
    indicating how similar a given estimate is to all the other estimates we've observed. `Standardizer` stores the
    transforms and population mean and variance for every parameter, allowing us to convert back and forth
    between natural space (:math:`rate`), transformed space (:math:`\text{ln}\; rate`), and standardized space
    (:math:`\left( \text{ln}\; rate  - \mu_{\text{ln}\; rate} \right)/\sigma_{\text{ln}\; rate}`).

    Typically, a :class:`Standardizer` will be constructed from a dataframe (:meth:`from_DataFrame`),
    but the individual means and variances can be provided at instantiation as well. Note, however,
    that these should be the mean/std of the *transformed* variable. For example, if `r` should be treated as
    log-normal with a natural-space mean of 1 and variance of 0.1, the right way to instantiate the class
    would be `Standardizer(d={'μ': 0, 'σ2': 0.1}, log_vars=['d'])`.


    Notes
    -----
    :class:`Standardizer` is just a `dictionary <https://docs.python.org/3/tutorial/datastructures.html#dictionaries>`_
    with some extra methods and defaults, so standard dictionary methods like :meth:`dict.update` still work.


    Parameters
    ----------
    log_vars: list, optional
        List of input and output variables to be treated as log-normal.
    logit_vars: list, optional
        List of input and output variables to be treated as logit-normal.
    **kwargs
        Mean and variance of each variable as a dictionary, e.g. d={'μ': 0, 'σ2': 0.1}


    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from gumbi import Standardizer
    >>> stdzr = Standardizer(x={'μ': 1, 'σ2': 0.1}, d={'μ': 0, 'σ2': 0.1}, log_vars=['d'])

    Transforming and standardizing a single parameter:

    >>> stdzr.transform('x', μ=1)
    1
    >>> stdzr.stdz('x', 1)
    0.0
    >>> stdzr.unstdz('x', 0)
    1.0
    >>> stdzr.stdz('x', 1+0.1**0.5)
    1.0  # approximately
    >>> stdzr.unstdz('x', 1)
    1.316227766016838
    >>> stdzr.stdz('d', 1)
    0.0
    >>> stdzr.stdz('d', np.exp(0.1**0.5))
    1.0  # approximately

    Transforming and standardizing a distribution:

    >>> stdzr.transform('x', μ=1., σ2=0.1)
    (1, 0.1)
    >>> stdzr.stdz('x', 1, 0.1)
    (0.0, 1.0)
    >>> stdzr.stdz('d', 1, 0.1)
    (0.0, 1.0)
    >>> stdzr.transform('d', 1, 0.1)
    (0.0, 0.1)

    Standardizing a series:

    >>> x_series = pd.Series(np.arange(1,5), name='x')
    >>> stdzr.stdz(x_series)
    0    0.000000
    1    3.162278
    2    6.324555
    3    9.486833
    Name: x, dtype: float64
    >>> r_series = pd.Series(np.arange(1,5), name='d')
    >>> stdzr.stdz(r_series)
    0    0.000000
    1    2.191924
    2    3.474117
    3    4.383848
    Name: r, dtype: float64

    """

    # TODO: Standardizer: make transform suggestions based on provided tidy? e.g., all>0 -> log/exp

    def __init__(self, log_vars=None, logit_vars=None, **kwargs):
        self.validate(kwargs)
        for name, stats in kwargs.items():
            if 'σ2' not in stats:
                stats['σ2'] = stats['σ']**2
                del stats['σ']
        super().__init__(**kwargs)
        self._transforms = {var: [skip, skip] for var in kwargs.keys()}
        if log_vars is not None:
            log_vars = [log_vars] if isinstance(log_vars, str) else log_vars
            if not isinstance(log_vars, list):
                raise TypeError('log_vars must be a list or str')
            self._transforms |= {var: [np.log, np.exp] for var in log_vars}
        if logit_vars is not None:
            logit_vars = [logit_vars] if isinstance(logit_vars, str) else logit_vars
            if not isinstance(logit_vars, list):
                raise TypeError('logit_vars must be a list or str')
            self._transforms |= {var: [logit, expit] for var in logit_vars}
        self._log_vars = log_vars if log_vars is not None else []
        self._logit_vars = logit_vars if logit_vars is not None else []

    def __or__(self, __dct) -> Standardizer:
        new_dct = super().__or__(__dct)
        stdzr = Standardizer(**new_dct)
        if isinstance(__dct, Standardizer):
            stdzr.transforms = self.transforms | __dct.transforms
        else:
            stdzr.transforms = self.transforms
        return stdzr

    def __ror__(self, __dct) -> Standardizer:
        new_dct = super().__ror__(__dct)
        stdzr = Standardizer(**new_dct)
        stdzr.transforms = self.transforms
        return stdzr

    def __repr__(self):
        summary = '\n\t'.join([
            f'Standardizer:',
            f'log_vars: {self.log_vars}',
            f'logit_vars: {self.logit_vars}',
        ]) + '\n\n' + str({**self})
        return summary

    @property
    def log_vars(self) -> list[str]:
        """List of log-normal variables"""
        return self._log_vars

    @log_vars.setter
    def log_vars(self, var_list):
        var_list = [var_list] if isinstance(var_list, str) else var_list
        if not isinstance(var_list, list):
            raise TypeError('log_vars must be a list or str')
        self._log_vars = var_list
        self._transforms |= {var: [np.log, np.exp] for var in var_list}

    @property
    def logit_vars(self) -> list[str]:
        """List of logit-normal variables"""
        return self._logit_vars

    @logit_vars.setter
    def logit_vars(self, var_list):
        var_list = [var_list] if isinstance(var_list, str) else var_list
        if not isinstance(var_list, list):
            raise TypeError('logit_vars must be a list or str')
        self._logit_vars = var_list
        self._transforms |= {var: [logit, expit] for var in var_list}

    @property
    def transforms(self) -> dict:
        """Collection of forward and reverse transform functions for each variable"""
        return self._transforms

    @transforms.setter
    def transforms(self, dct) -> dict:
        self._transforms = dct
        self._log_vars = [v for v, lst in dct.items() if lst[0] is np.log]
        self._logit_vars = [v for v, lst in dct.items() if lst[0] is logit]

    @classmethod
    def validate(cls, dct: dict):
        """Ensures provided dictionary has all required attributes"""
        assert all('μ' in sub.keys() for sub in dct.values())
        assert all(('σ' in sub.keys() or 'σ2' in sub.keys()) for sub in dct.values())

    @classmethod
    def from_DataFrame(cls, df: pd.DataFrame, log_vars=None, logit_vars=None):
        """Construct from wide-form DataFrame"""
        float_columns = df.dtypes[df.dtypes == 'float64'].index.to_list()

        new = cls(log_vars=log_vars, logit_vars=logit_vars)

        dct = (df[float_columns]
               .apply(new.transform)
               .agg([np.mean, np.var])
               .rename(index={"mean": "μ", "var": "σ2"})
               .to_dict()
               )

        return new | dct

    def transform(self, name: str | pd.Series, μ: float = None, σ2: float = None) -> float | tuple | pd.Series:
        """Transforms a parameter, distribution, or Series

        Parameters
        ----------
        name: str or pd.Series
            Name of parameter. If a Series is supplied, the name of the series must be the parameter name.
        μ: float, optional
            Value of parameter or mean of parameter distribution. Only optional if first argument is a Series.
        σ2: float, optional
            Variance of parameter distribution.

        Returns
        -------
        float, tuple, or pd.Series
            Transformed parameter, (mean, variance) of untransformed distribution, or untransformed Series
        """
        if isinstance(series := name, pd.Series):
            return self._transform_value(series.name, series)
        elif μ is None:
            raise ValueError('μ cannot be None')
        if σ2 is None:
            return self._transform_value(name, μ)
        else:
            return self._transform_dist(name, μ, σ2)

    def untransform(self, name: str | pd.Series, μ: float = None, σ2: float = None) -> float | tuple | pd.Series:
        """Untransforms a parameter, distribution, or Series

        Parameters
        ----------
        name: str or pd.Series
            Name of parameter. If a Series is supplied, the name of the series must be the parameter name.
        μ: float, optional
            Value of parameter or mean of parameter distribution. Only optional if first argument is a Series.
        σ2: float, optional
            Variance of parameter distribution.

        Returns
        -------
        float, tuple, or pd.Series
            Untransformed parameter, (mean, variance) of untransformed distribution, or untransformed Series
        """
        if isinstance(series := name, pd.Series):
            return self._untransform_value(series.name, series)
        if σ2 is None:
            return self._untransform_value(name, μ)
        else:
            return self._untransform_dist(name, μ, σ2)

    def stdz(self, name: str | pd.Series, μ: float = None, σ2: float = None) -> float | tuple | pd.Series:
        """Transforms, mean-centers, and scales a parameter, distribution, or Series

        Parameters
        ----------
        name: str or pd.Series
            Name of parameter. If a Series is supplied, the name of the series must be the parameter name.
        μ: float, optional
            Value of parameter or mean of parameter distribution. Only optional if first argument is a Series.
        σ2: float, optional
            Variance of parameter distribution.

        Returns
        -------
        float, tuple, or pd.Series
            Standardized parameter, (mean, variance) of standardized distribution, or standardized Series
        """

        if isinstance(series := name, pd.Series):
            return self._stdz_value(series.name, series)
        if σ2 is None:
            return self._stdz_value(name, μ)
        else:
            return self._stdz_dist(name, μ, σ2)

    def unstdz(self, name: str | pd.Series, μ: float = None, σ2: float = None) -> float | tuple | pd.Series:
        """Untransforms, un-centers, and un-scales a parameter, distribution, or Series

        Parameters
        ----------
        name: str or pd.Series
            Name of parameter. If a Series is supplied, the name of the series must be the parameter name.
        μ: float, optional
            Value of parameter or mean of parameter distribution. Only optional if first argument is a Series.
        σ2: float, optional
            Variance of parameter distribution.

        Returns
        -------
        float, tuple, or pd.Series
            Unstandardized parameter, (mean, variance) of unstandardized distribution, or unstandardized Series
        """

        if isinstance(series := name, pd.Series):
            return self._unstdz_value(series.name, series)
        if σ2 is None:
            return self._unstdz_value(name, μ)
        else:
            return self._unstdz_dist(name, μ, σ2)

    def _transform_value(self, name: str, x: float) -> float:
        ftransform = self.transforms.get(name, [skip, skip])[0]

        return ftransform(x)

    def _untransform_value(self, name: str, x: float) -> float:
        rtransform = self.transforms.get(name, [skip, skip])[1]
        x_ = rtransform(x)

        return x_

    def _stdz_value(self, name: str, x: float) -> float:
        x_ = self.transform(name, x)
        μ = self.get(name, {'μ': 0})['μ']
        σ2 = self.get(name, {'σ2': 1})['σ2']
        σ = np.sqrt(σ2)
        return (x_ - μ) / σ

    def _unstdz_value(self, name: str, z: float) -> float:
        μ = self.get(name, {'μ': 0})['μ']
        σ2 = self.get(name, {'σ2': 1})['σ2']
        σ = np.sqrt(σ2)
        x_ = z * σ + μ
        return self.untransform(name, x_)

    @property
    def mean_transforms(self):
        """Function that transforms the mean of a distribution.

        These transform's should follow scipy's conventions such that a distribution can be defined in the given
        space by passing (loc=μ, scale=σ2**0.5). For a lognormal variable, an RV defined as ``lognorm(loc=μ,
        scale=σ2**0.5)`` in "natural" space is equivalent to ``norm(loc=np.log(μ), scale=σ2**0.5)`` in log space,
        so this transform should return ``np.log(μ)`` when converting from natural to log space, and ``np.exp(μ)``
        when converting from log to natural space. Similarly for a logit-normal variable, an RV defined as
        ``logitnorm(loc=μ, scale=σ2**0.5))`` in natural space is equivalent to ``norm(loc=logit(μ), scale=σ2**0.5)``
        in logit space, so this transform should return ``logit(μ)`` when converting from natural to logit space,
        and ``expit(μ)`` when converting from logit to natural space.
        """

        # Forward and reverse transform for each variable type
        transforms = {skip: [lambda μ, σ2: μ,
                             lambda μ, σ2: μ],
                      # Note these are no longer strictly mean and variance. They are defined to be compatible with
                      # scipy.stats.lognormal definition
                      np.log: [lambda μ, σ2: np.log(μ),
                               lambda μ, σ2: np.exp(μ)],
                      logit: [lambda μ, σ2: logit(μ),
                              lambda μ, σ2: expit(μ)]
                      }
        return transforms

    @property
    def var_transforms(self):
        """Function that transforms the variance of a distribution.

        These transform's should follow scipy's conventions such that a distribution can be defined in the given
        space by passing (loc=μ, scale=σ2**0.5). Accordingly, since both log-normal and logit-normal variables are
        defined in terms of the scale (standard deviation) in their respective transformed spaces, this function
        simply returns the variance unchanged in these cases.
        """

        # Forward and reverse transform for each variable type
        transforms = {skip: [lambda μ, σ2: σ2,
                             lambda μ, σ2: σ2],
                      np.log: [lambda μ, σ2: σ2,
                               lambda μ, σ2: σ2],
                      logit: [lambda μ, σ2: σ2,
                              lambda μ, σ2: σ2]
                      }
        return transforms

    def _transform_dist(self, name: str, mean: float, var: float) -> tuple:
        f_transform = self.transforms.get(name, [skip, skip])[0]
        f_mean_transform = self.mean_transforms[f_transform][0]
        f_var_transform = self.var_transforms[f_transform][0]

        mean_ = f_mean_transform(mean, var)
        var_ = f_var_transform(mean, var)
        return mean_, var_

    def _untransform_dist(self, name: str, mean: float, var: float) -> tuple:
        f_transform = self.transforms.get(name, [skip, skip])[0]
        r_mean_transform = self.mean_transforms[f_transform][1]
        r_var_transform = self.var_transforms[f_transform][1]

        mean_ = r_mean_transform(mean, var)
        var_ = r_var_transform(mean, var)

        return mean_, var_

    def _stdz_dist(self, name: str, mean: float, var: float) -> tuple:
        mean_, var_ = self.transform(name, mean, var)
        μ = self.get(name, {'μ': 0})['μ']
        σ2 = self.get(name, {'σ2': 1})['σ2']
        σ = np.sqrt(σ2)
        mean_z = (mean_ - μ) / σ
        var_z = var_ / σ2
        return mean_z, var_z

    def _unstdz_dist(self, name: str, z_mean: float, z_var: float) -> tuple:
        μ = self.get(name, {'μ': 0})['μ']
        σ2 = self.get(name, {'σ2': 1})['σ2']
        σ = np.sqrt(σ2)
        mean_ = z_mean * σ + μ
        var_ = z_var * σ2
        mean, var = self.untransform(name, mean_, var_)
        return mean, var


@dataclass
class MetaFrame(pd.DataFrame, ABC):
    """Abstract Base Class for :class:`WideData` and :class:`TidyData`."""

    df: pd.DataFrame
    outputs: list
    log_vars: list = None
    logit_vars: list = None
    names_column: str = 'Variable'
    values_column: str = 'Value'
    stdzr: Standardizer = None

    _metadata = ['df', 'outputs', 'log_vars', 'logit_vars', 'names_column', 'values_column', 'stdzr']

    def __post_init__(self):
        super(MetaFrame, self).__init__(self.df)
        if self.stdzr is None:
            self.stdzr = Standardizer.from_DataFrame(self.df, log_vars=self.log_vars, logit_vars=self.logit_vars)
        else:
            self.log_vars = self.stdzr.log_vars
            self.logit_vars = self.stdzr.logit_vars
        del self.df

    def __repr__(self):
        cls = self.__class__.__name__
        df_repr = super(MetaFrame, self).__repr__()

        summary = '\n\t'.join([
            f'{cls}:',
            f'outputs: {self.outputs}',
            f'inputs: {self.inputs}',
        ]) + '\n\n' + df_repr
        return summary

    @property
    @abstractmethod
    def z(self) -> pd.DataFrame:
        """Standardized data values."""
        pass

    @property
    @abstractmethod
    def t(self) -> pd.DataFrame:
        """Transformed data values."""
        pass

    @property
    def specs(self) -> dict:
        """Provides keyword arguments for easy instantiation of a similar object."""
        return dict(outputs=self.outputs, names_column=self.names_column, values_column=self.values_column,
                    stdzr=self.stdzr, log_vars=self.log_vars, logit_vars=self.logit_vars)

    @property
    def inputs(self) -> list[str]:
        """Columns of dataframe not contained in :attr:`outputs`."""
        return [col for col in self.columns if col not in self.outputs]

    @property
    def float_inputs(self) -> list[str]:
        """Columns of dataframe with "float64" dtype."""
        return [col for col in self.inputs if self[col].dtype == 'float64']

    @classmethod
    def _wide_to_tidy_(cls, wide, outputs, names_column='Variable', values_column='Value'):
        inputs = [col for col in wide.columns if col not in outputs]
        tidy = wide.melt(id_vars=inputs, value_vars=outputs, var_name=names_column,
                         value_name=values_column)
        return tidy

    @classmethod
    def _tidy_to_wide_(cls, tidy, names_column='Variable', values_column='Value'):
        inputs = [col for col in tidy.columns if col not in [names_column, values_column]]
        wide = (tidy
                .pivot(index=inputs, columns=names_column, values=values_column)
                .reset_index()
                .rename_axis(columns=None)
                )
        return wide


class WideData(MetaFrame):
    """Container for wide-form tabular data, allowing simple access to standardized and/or transformed values.

    Note that :class:`WideData` is instantiated with a **wide-form** dataframe. This class is not intended to be
    instantiated directly, use :class:`DataSet` instead. :class:`WideData` subclasses pandas' DataFrame,
    which everyone says is a bad idea, so be prepared for unexpected behavior if instantiated directly. Namely, in-place
    modifications return a :class:`WideData` type correctly, but slices return a `pd.DataFrame` type.

    Parameters
    ----------
    data: pd.DataFrame
        A wide-form dataframe.
    outputs: list
        Columns of `data` to be treated as outputs.
    names_column: str, default 'Variable'
        Name to be used in tidy view for column containing output names.
    values_column: str, default 'Value'
        Name to be used in tidy view for column containing output values.
    log_vars: list, optional
        List of input and output variables to be treated as log-normal. Ignored if `stdzr` is supplied.
    logit_vars: list, optional
        List of input and output variables to be treated as logit-normal. Ignored if `stdzr` is supplied.
    stdzr: Standardizer, optional
        An :class:`Standardizer` instance. If not supplied, one will be created automatically.
    """

    @property
    def z(self) -> pd.DataFrame:
        """Standardized data values."""
        df_ = self.copy()
        cols = self.outputs + self.float_inputs
        df_[cols] = df_[cols].apply(self.stdzr.stdz)
        return df_

    @property
    def t(self) -> pd.DataFrame:
        """Transformed data values."""
        df_ = self.copy()
        cols = self.outputs + self.float_inputs
        df_[cols] = df_[cols].apply(self.stdzr.transform)
        return df_

    def to_tidy(self) -> TidyData:
        """Converts to TidyData"""
        tidy = TidyData(self, **self.specs)
        return tidy

    @classmethod
    def from_tidy(cls, tidy, outputs=None, names_column='Variable', values_column='Value',
                  stdzr=None, log_vars=None, logit_vars=None):
        """Constructs `WideData` from a tidy-form dataframe. See :class:`WideData` for explanation of arguments."""
        outputs = outputs if outputs is not None else list(tidy[names_column].unique())
        wide = cls._tidy_to_wide_(tidy, names_column=names_column, values_column=values_column)
        return cls(wide, outputs=outputs, names_column=names_column, values_column=values_column,
                   stdzr=stdzr, log_vars=log_vars, logit_vars=logit_vars)


class TidyData(MetaFrame):
    """Container for tidy-form tabular data, allowing simple access to standardized and/or transformed values.

    Note that :class:`TidyData` is instantiated with a **wide-form** dataframe. This class is not intended to be
    instantiated directly, use :class:`DataSet` instead. :class:`TidyData` subclasses pandas' DataFrame,
    which everyone says is a bad idea, so be prepared for unexpected behavior if instantiated directly. Namely, in-place
    modifications return a :class:`TidyData` type correctly, but slices return a `pd.DataFrame` type.

    Parameters
    ----------
    data: pd.DataFrame
        A wide-form dataframe.
    outputs: list
        Columns of `data` to be treated as outputs.
    names_column: str, default 'Variable'
        Name to be used in tidy view for column containing output names.
    values_column: str, default 'Value'
        Name to be used in tidy view for column containing output values.
    log_vars: list, optional
        List of input and output variables to be treated as log-normal. Ignored if `stdzr` is supplied.
    logit_vars: list, optional
        List of input and output variables to be treated as logit-normal. Ignored if `stdzr` is supplied.
    stdzr: Standardizer, optional
        An :class:`Standardizer` instance. If not supplied, one will be created automatically.
    """

    def __post_init__(self):
        tidy = self._wide_to_tidy_(self.df, outputs=self.outputs, names_column=self.names_column,
                                   values_column=self.values_column)
        self.df = tidy
        super(TidyData, self).__post_init__()

    @property
    def z(self) -> pd.DataFrame:
        """Standardized data values."""
        wide = self.to_wide()
        specs = dict(outputs=self.outputs, names_column=self.names_column, values_column=self.values_column)
        wd = WideData(wide, **specs, stdzr=self.stdzr)
        z = self._wide_to_tidy_(wd.z, **specs)
        return z

    @property
    def t(self) -> pd.DataFrame:
        """Transformed data values."""
        wide = self.to_wide()
        specs = dict(outputs=self.outputs, names_column=self.names_column, values_column=self.values_column)
        wd = WideData(wide, **specs, stdzr=self.stdzr)
        z = self._wide_to_tidy_(wd.t, **specs)
        return z

    def to_wide(self) -> WideData:
        """Converts to WideData"""
        wide_df = self._tidy_to_wide_(self, names_column=self.names_column, values_column=self.values_column)
        wide = WideData(wide_df, **self.specs)
        return wide


@dataclass
class DataSet:
    """Container for tabular data, allowing simple access to standardized values and wide or tidy dataframe formats.

    :class:`DataSet` is instantiated with a **wide-form** dataframe, with all outputs of a given observation in a
    single row, but allows easy access to the corresponding **tidy** dataframe, with each output in a separate row (
    the :meth:`from_tidy` also allows construction from tidy data`). The titles of the tidy-form columns for the
    output names and their values are supplied at instantiation, defaulting to "Variable" and "Value". For example,
    say we have an observation at position (x,y) with measurements of i, j, and k. The wide-form dataframe would have
    one column for each of x, y, i, j, and k, while the tidy-form dataframe would have a column for each of x and y,
    a "Variable" column where each row contains either "i", "j", or "k" as strings, and a "Value" column containing
    the corresponding measurement. Wide data is more space-efficient and perhaps more intuitive to construct and
    inspect, while tidy data more clearly distinguishes inputs and outputs. These views are accessible through the
    :attr:`wide` and :attr:`tidy` attributes as instances of :class:`WideData` and :class:`TidyData`, respectively.

    As a container for :class:`WideData` and :class:`TidyData`, this class also provides simple access to
    standardized values of the data through `wide.z` and `tidy.z` or transformed values through `wide.t` and
    `tidy.t`. A :class:`Standardizer` instance can be supplied as a keyword argument, otherwise one will be
    constructed automatically from the supplied dataframe with the supplied values of `log_vars` and `logit_vars`.
    Unlike :class:`WideData` and :class:`TidyData`, the :attr:`wide` and :attr:`tidy` attributes of a *DataSet* can
    be altered and sliced while retaining their functionality, with a cursory integrity check. The
    :class:`Standardizer` instance can be updated with :meth:`update_stdzr`, for example following manipulation of
    the data or alteration of :attr:`log_vars` and :attr:`logit_vars`.

    Parameters
    ----------
    data: pd.DataFrame
        A wide-form dataframe. See class method :meth:`from_tidy` for instantiation from tidy data.
    outputs: list
        Columns of `data` to be treated as outputs.
    names_column: str, default 'Variable'
        Name to be used in tidy view for column containing output names.
    values_column: str, default 'Value'
        Name to be used in tidy view for column containing output values.
    log_vars: list, optional
        List of input and output variables to be treated as log-normal. Ignored if `stdzr` is supplied.
    logit_vars: list, optional
        List of input and output variables to be treated as logit-normal. Ignored if `stdzr` is supplied.
    stdzr: Standardizer, optional
        An :class:`Standardizer` instance. If not supplied, one will be created automatically.

    Examples
    --------
    >>> df = pd.read_pickle(test_data / 'estimates_test_data.pkl')
    >>> ds = DataSet.from_tidy(df, names_column='Parameter', log_vars=['Y', 'c', 'b'], logit_vars=['X', 'e'])
    >>> ds
    DataSet:
        wide: [66 rows x 13 columns]
        tidy: [396 rows x 9 columns]
        outputs: ['e', 'f', 'b', 'c', 'a', 'd']
        inputs: ['Code', 'Target', 'Y', 'X', 'Reaction', 'lg10_Z', 'Metric']

    >>> ds.wide = ds.wide.drop(range(0,42,2))
    DataSet:
        wide: [45 rows x 13 columns]
        tidy: [270 rows x 9 columns]
        outputs: ['e', 'f', 'b', 'c', 'a', 'd']
        inputs: ['Code', 'Target', 'Y', 'X', 'Reaction', 'lg10_Z', 'Metric']

    >>> ds.tidy.z  # tidy-form dataframe with standardized values
    >>> ds.wide.z  # wide-form dataframe with standardized values
    """

    data: pd.DataFrame
    outputs: list
    names_column: str = 'Variable'
    values_column: str = 'Value'
    log_vars: list = None
    logit_vars: list = None
    stdzr: Standardizer = None

    def __post_init__(self):
        if self.stdzr is None:
            self.stdzr = Standardizer.from_DataFrame(self.wide, log_vars=self.log_vars, logit_vars=self.logit_vars)
        else:
            self.log_vars = self.stdzr.log_vars
            self.logit_vars = self.stdzr.logit_vars

    def __repr__(self):
        wide_shape = '[{0} rows x {1} columns]'.format(*self.wide.shape)
        tidy_shape = '[{0} rows x {1} columns]'.format(*self.tidy.shape)
        summary = '\n\t'.join([
            'DataSet:',
            f'wide: {wide_shape}',
            f'tidy: {tidy_shape}',
            f'outputs: {self.outputs}',
            f'inputs: {self.inputs}',
        ])
        return summary

    @property
    def specs(self):
        """Provides keyword arguments for easy instantiation of a similar :class:`DataSet`."""
        return dict(outputs=self.outputs, names_column=self.names_column, values_column=self.values_column,
                    stdzr=self.stdzr, log_vars=self.log_vars, logit_vars=self.logit_vars)

    @property
    def inputs(self):
        """Columns of dataframe not contained in :attr:`outputs`."""
        return [col for col in self.wide.columns if col not in self.outputs]

    @property
    def float_inputs(self):
        """Columns of dataframe with "float64" dtype."""
        return [col for col in self.inputs if self.wide[col].dtype == 'float64']

    @property
    def wide(self) -> WideData:
        """Wide-form view of data"""
        return WideData(self.data, **self.specs)

    @wide.setter
    def wide(self, wide_df: pd.DataFrame):
        assert any([output in wide_df.columns for output in self.outputs]), \
            f'Dataframe must have at least one of outputs {self.outputs}'
        self.data = wide_df

    @property
    def tidy(self) -> TidyData:
        """Tidy-form view of data"""
        return TidyData(self.data, **self.specs)

    @tidy.setter
    def tidy(self, tidy_df: pd.DataFrame):
        assert all([col in tidy_df.columns for col in [self.names_column, self.values_column]]), \
            f'Dataframe must have both columns {[self.names_column, self.values_column]}'
        self.wide = WideData.from_tidy(tidy_df, **self.specs)

    @classmethod
    def from_tidy(cls, tidy, outputs=None, names_column='Variable', values_column='Value',
                  stdzr=None, log_vars=None, logit_vars=None):
        """Constructs a `DataSet` from a tidy-form dataframe. See :class:`DataSet` for explanation of arguments."""
        assert all([col in tidy.columns for col in [names_column, values_column]]), \
            f'Dataframe must have both columns {[names_column, values_column]}'
        specs = dict(outputs=outputs, names_column=names_column, values_column=values_column,
                     stdzr=stdzr, log_vars=log_vars, logit_vars=logit_vars)
        wide = WideData.from_tidy(tidy, **specs)
        return cls(wide, **wide.specs)

    @classmethod
    def from_wide(cls, wide, outputs=None, names_column='Variable', values_column='Value',
                  stdzr=None, log_vars=None, logit_vars=None):
        """Constructs a `DataSet` from a wide-form dataframe. See :class:`DataSet` for explanation of arguments."""
        return cls(wide, outputs=outputs, names_column=names_column, values_column=values_column, stdzr=stdzr,
                   log_vars=log_vars, logit_vars=logit_vars)

    def update_stdzr(self):
        """Updates internal :class:`Standardizer` with current data, :attr:`log_vars`, and :attr:`logit_vars`."""
        self.stdzr |= Standardizer.from_DataFrame(self.wide, log_vars=self.log_vars, logit_vars=self.logit_vars)
